
/* BEGIN COPYRIGHT NOTICE

Copyright (c) 1992 by John Sarraille and Peter DiFalco
(john@ishi.csustan.edu)

Copyright (c) 1998 by John Sarraille
(john@ishi.csustan.edu)

Permission to use, copy, modify, and distribute this software
and its documentation for any purpose and without fee is hereby
granted, provided that the above copyright notice appear in all
copies and that both that copyright notice and this permission
notice appear in supporting documentation.

The algorithm used in this program was inspired by the paper
entitled "A Fast Algorithm To Determine Fractal Dimensions By
Box Counting", which was written by Liebovitch and Toth, and
which appeared in the journal "Physics Letters A", volume 141,
pp 386-390, (1989).

This program is not warranted: use at your own risk.

END COPYRIGHT NOTICE */


/*
   FD3 PROGRAM FILE

Filename:  datarep.c

OUTLINE OF CALLS:

GetNormztnInfo +
  ZeroDiffsErr +
  AllocMM +
    gmalloc
  GetMaxMin +
    Gfopen 
    GfscanfULI
    InitMaxMin +
      GfscanfDbl
    ScanMaxMin +
      gFSCANFdBL
  GetMaxDiff +
    MaxDiffOflowErr +
  NoScaleErr +
  PrintNormztnInfo +
GetRAMsize +
GetRAM +
  gmalloc
InitRAM +
  CreateQ
LoadRAM +
  InitLR +
    Gfopen 
    gFSCANFdBL
  LRScan +
    gFSCANFdBL
    ENQ
  FinLR +
    PrintRAMlists +
      PrintQ
SortRAM +
  SRstartMsg +
  RadPass +      (replaced by macro rADpASS)
    SizeQ
    ReQsrcRun +  (replaced by macro REqSRcRUN)
      DEQ
      ISa
      ENQ
    ReQgenCase + (replaced by macro REqGENcASE)
      DEQfAST
      ISa
      ENQfAST
  ConcatQs
  SRfinMsg +
    PrintRAMlists +
    PrintRAMmesh +
      PrintQmesh
*/

#include "fd.h"
#include "file.h"
#include "qmacro.h"
#include "sort.h"

/*
      THE FUNCTIONS IN THIS FILE
*/
void  GetNormztnInfo         () ;
void  ZeroDiffsErr           () ;
void  AllocMM                () ;
void  GetMaxMin              () ;
void  InitMaxMin             () ;
void  ScanMaxMin             () ;
void  GetMaxDiff             () ;
void  MaxDiffOflowErr        () ;
void  NoScaleErr             () ;
void  PrintNormztnInfo       () ;
void  GetRAMsize             () ;
void  GetRAM                 () ;
void  InitRAM                () ;
void  LoadRAM                () ;
void  InitLR                 () ;
void  LRScan                 () ;
void  FinLR                  () ;
void  PrintRAMlists          () ;
void  SortRAM                () ;
void  SRstartMsg             () ;
void  RadPass                () ;
void  ReQsrcRun              () ;
void  ReQgenCase             () ;
void  SRfinMsg               () ;
void  PrintRAMmesh           () ;

/*
     THE FUNCTIONS NOT IN THIS FILE, BUT CALLED FROM HERE
*/
extern void * gmalloc    () ;    /*  in mem.c    */
extern void  Gfopen      () ;    /*  in file.c   */
extern void  GfscanfULI  () ;    /*  in file.c   */
extern void  GfscanfDbl  () ;    /*  in file.c   */
extern void  CreateQ     () ;    /*  in queue.c  */
extern void  PrintQ      () ;    /*  in queue.c  */
extern uLong SizeQ       () ;    /*  in queue.c  */
extern void  ConcatQs    () ;    /*  in queue.c  */
extern void  PrintQmesh  () ;    /*  in queue.c  */

/*
           MACROS CALLED FROM THIS FILE

    gFSCANFdBL         (*  in file.h    *)
    ENQ                (*  in qmacro.h  *)
    rADpASS            (*  in sort.h    *)
    REqSRcRUN          (*  in sort.h    *)
    DEQ                (*  in qmacro.h  *)
    ISa                (*  in fd.h      *)
    ENQ                (*  in qmacro.h  *)
    REqGENcASE         (*  in sort.h    *)
    DEQfAST            (*  in qmacro.h  *)
    ENQfAST            (*  in qmacro.h  *)
*/

/*
     CALLS TO THESE FUNCTIONS HAVE BEEN REPLACED BY MACRO CALLS.
*/
extern void  Deq         () ;    /*  in queue.c  */
extern void  DeqFast     () ;    /*  in queue.c  */
extern void  Enq         () ;    /*  in queue.c  */
extern void  EnqFast     () ;    /*  in queue.c  */

/* ######################################################### */
/*                      GetNormztnInfo                       */
/* ######################################################### */
/*
     LOAD A STRUCTURE WITH ALL THE INFORMATION THAT WILL BE
     NEEDED TO PERFORM THE NORMALIZATION OF THE DATA.  DATA
     WILL BE MOVED TO THE ORIGIN, AND SCALED SO THAT THE
     LARGEST COORDINATE VALUE EQUALS (NEARLY) THE LARGEST
     UNSIGNED LONG INTEGER REPRESENTABLE.

     WE DID NOT HAVE TO MODIFY THIS FUNCTION TO GET THE
     mincover OPTION WORKING.  HOWEVER, CHANGES MAY BE
     NECESSARY WHEN AND IF THE mincover OPTION IS UPGRADED TO
     HANDLE INPUTS WITH EMBEDDING DIMENSION GREATER THAN 1.
*/
void GetNormztnInfo (commandPtr, machinePtr, fileRepPtr, normInfoPtr)
        commandType    * commandPtr;
        machineType    * machinePtr;
        fileRepType    * fileRepPtr;
        normInfoType   * normInfoPtr ;
{
       /* temp floating point estimates of maxCoord */
   double maxCtemp, flrTemp ;
   uLong maxUlong = (machinePtr->maxUlong) ;
   
   (normInfoPtr->ulongBitNum) = (machinePtr->ulongBitNum) ;
   (normInfoPtr->maxUlong) = maxUlong ;
   (normInfoPtr->embedDim) = (fileRepPtr->embedDim) ;
   AllocMM (commandPtr, machinePtr, normInfoPtr) ;

   if (!(commandPtr->terse))
      { printf("GetNormztnInfo: Finding max's and min's ... ") ;
        fflush(stdout); }
   GetMaxMin (fileRepPtr, normInfoPtr, machinePtr) ;
   if (!(commandPtr->terse)) { printf("Done.\n") ; fflush(stdout); }

   GetMaxDiff (normInfoPtr, machinePtr->maxDouble) ;
      /*
          HERE WE MAKE AN ATTEMPT TO DEAL WITH THE PROBLEM OF
          THE MAXDIFF BEING ZERO.  IT'S DOUBTFUL THAT THIS
          SOLVES *ALL* POSSIBLE PROBLEMS RELATED TO MAXDIFF
          BEING *CLOSE* TO ZERO.  IF MAXDIFF IS VERY SMALL, BUT
          NON-ZERO, WE COULD GET A VALUE FOR THE SCALING FACTOR
          THAT IS VERY MUCH IN ERROR.
      */ 
   if ((normInfoPtr->maxDiff) == 0) ZeroDiffsErr(); /* THIS ABORTS */
   else (normInfoPtr->scalingFactor) =
          fabs(maxUlong/(normInfoPtr->maxDiff));
      /*
         HERE WE ARE CONCERNED WITH CALCULATING maxCoord, THE
         LARGEST VALUE WE CAN EXPECT TO GET WHEN WE USE
         scalingFactor TO SCALE THE DATA.  DUE TO THE LACK OF
         PRECISION INHERENT IN DIVIDING BY maxDiff ABOVE,
         maxCoord COULD BE QUITE DIFFERENT FROM maxUlong.  WE
         DON'T ALLOW THE PROGRAM TO CONTINUE IF maxCtemp (AN
         APPROXIMATION TO maxCoord) DIFFERS FROM maxUlong BY
         10% OR MORE.  WE FIGURE THAT IF THE VALUE IS OFF BY
         THAT MUCH THEN WE MAY HAVE LOST SO MUCH PRECISION THAT
         IT'S MORE PRUDENT TO WARN THE USER BY QUITTING THAN TO
         CONTINUE.

         AT THIS TIME, MaxCoord IS USED ONLY TWO WAYS:  FIRST,
         IT IS USED HERE TO DECIDE WHETHER TO ABORT THE
         PROGRAM, AND SECOND IT IS USED WITH THE -m (mincover)
         OPTION, TO DETERMINE THE LARGEST CELL SIZE.  
      */
   maxCtemp = (normInfoPtr->maxDiff) * (normInfoPtr->scalingFactor) ;
   if (       (maxCtemp >= (1.10*maxUlong) )
           || (maxCtemp <= (0.90*maxUlong) )      )
      NoScaleErr() ; /* THIS ABORTS. */
      /*
         IT'S POSSIBLE FOR maxCtemp TO BE LARGER THAN
         maxUlong.  HERE, WE CAP IT.
      */
   if   (maxCtemp >= maxUlong)
        (normInfoPtr->maxCoord) = maxUlong ;
   else
   {
       /*
          THIS PART OF THE CODE IS REACHED ONLY IF maxCtemp IS
          IN THE INTERVAL (0,maxUlong).  THUS IT SHOULD BE SAFE
          TO CONVERT IT TO A uLong.
       */
     flrTemp  = floor(maxCtemp) ;
     if   ( (maxCtemp - flrTemp) >= 0.5 )
          (normInfoPtr->maxCoord) = ((uLong)flrTemp) + 1 ;
     else (normInfoPtr->maxCoord) = ((uLong)flrTemp) ;
   }
   if ((commandPtr->debugging) || (commandPtr->scaleInfo))
         PrintNormztnInfo (normInfoPtr) ;
   return ;
}
/* ######################################################### */
/*                          ZeroDiffsErr                     */
/* ######################################################### */
void ZeroDiffsErr ()
   {  
      printf("\n");
      printf("PROBLEM:\n");
      printf("\n");
      printf("Function \"GetMaxDiff\" failed to find any differences\n");
      printf("among the points(s) in the input file.  Fd3 cannot \n");
      printf("continue under these circumstances.  The only possible\n");
      printf("estimate of the dimension of the set is 0,  because the \n");
      printf("set appears to consist of just one point!\n");
      printf("\n");
      printf("Sorry, fd3 must abort now.\n");
      printf("\n");
      fflush(stdout) ; exit(EXIT_FAILURE);
   }
/* ######################################################### */
/*                            AllocMM                        */
/* ######################################################### */
/*
    THIS IS A HELPER FUNCTION CALLED THAT MAKES SOME
    LOWER-LEVEL CALLS FOR THE FUNCTION GetNormztnInfo.  It
    ALLOCATES MEMORY NEEDED FOR A PAIR OF ARRAYS.  THE ARRAYS
    ARE USED TO HOLD MAXIMUM AND MINIMUM COORDINATE VALUES OF
    THE DATA.

     USE THIS STATEMENT TO TEST THE FAILURE HANDLING :-)
            (normInfoPtr->min) = (double *) gmalloc ("AllocMM",
                    "normInfoPtr->min", 1000000000000000000000 ) ;
*/
void  AllocMM (commandPtr, machinePtr, normInfoPtr) 
        commandType    * commandPtr;
        machineType    * machinePtr;
        normInfoType   * normInfoPtr ;
{
  size_t eD = (size_t) (normInfoPtr->embedDim), 
         dBN = (size_t) (machinePtr->doubleByteNum) ;

  if ( (commandPtr->debugging) || (commandPtr->checking) )
     { printf("AllocMM: Allocating storage for min and max coords ... ");
       fflush(stdout);  }

   (normInfoPtr->min) =
      (double *) gmalloc ("AllocMM", "normInfoPtr->min", eD * dBN) ;
   (normInfoPtr->max) =
       (double *) gmalloc ("AllocMM", "normInfoPtr->max", eD * dBN) ; 

  if ( (commandPtr->debugging) || (commandPtr->checking) )
      { printf("Done.\n\n"); fflush(stdout);  }

  return ;
}
/* ######################################################### */
/*                          GetMaxMin                        */
/* ######################################################### */
/*
    THIS FUNCTION OPENS THE INPUT FILE, SKIPS OVER THE FIRST
    NUMBER IN THE FILE (DataLines), INITIALIZES MAX AND MIN
    VALUES FOR EACH COORDINATE SLOT, FINDS THE MAX AND MIN OF
    EACH COORDINATE BY GOING THROUGH THE REST OF THE FILE,
    UPDATING max's AND min's, AND THEN CLOSES THE INPUT FILE.
    
    FOR EXAMPLE, IF THE POINTS LISTED IN THE INPUT FILE HAVE X,
    Y, AND Z COORDINATES, THEN THIS FUNCTION CALCULATES 6
    QUANTITIES:  THE MINIMUM X, Y, AND Z COORDINATES AND THE
    MAXIMUM X, Y, AND Z COORDINATES.  SINCE THE FUNCTION READS
    THE ENTIRE FILE, WE USE A MACRO TO DO THE READING, TO AVOID
    THE OVERHEAD OF A FUNCTION CALL.  *WARNING* TO THOSE WHO
    MAKE CHANGES TO THIS CODE:  DON'T CHANGE *ANYTHING* IN
    MACRO CALLS OR DEFINITIONS, UNLESS YOU ARE QUITE SURE YOU
    KNOW WHAT YOU ARE DOING.
*/
void GetMaxMin (fileRepPtr, nrmInfPtr, mPtr)
        fileRepType    * fileRepPtr;
        normInfoType   * nrmInfPtr ;
        machineType    * mPtr      ;
{ 
  FILE  * fVarPt ;
  char  * fName = (fileRepPtr->fileName) ;
  uLong   tempU ;

  Gfopen ("GetMaxMin", &fVarPt, fName, "rb+");
  GfscanfULI ("GetMaxMin", fName, (uLong) 1, fVarPt, &tempU, mPtr->maxUlong);
  InitMaxMin (fileRepPtr, nrmInfPtr, fVarPt, mPtr->maxDouble) ;
  ScanMaxMin (fileRepPtr, nrmInfPtr, fVarPt, mPtr->maxDouble) ;
  fclose (fVarPt) ;

  return ;
}
/* ######################################################### */
/*                         InitMaxMin                        */
/* ######################################################### */
void  InitMaxMin (frPtr, niPtr, fvPtr, maxD) 
        fileRepType    * frPtr;
        normInfoType   * niPtr ;
        FILE           * fvPtr  ;
        double         maxD ;
{
   uLong  col ;
   double tempD ;

   for (col = 0; col < (frPtr->embedDim); col++)
   { 
      GfscanfDbl ("InitMaxMin", (frPtr->fileName), (uLong) 2,
                    fvPtr, &tempD, maxD);
      (niPtr->max)[col] = tempD ;  
      (niPtr->min)[col] = tempD ;  
   }
   return ;
}
/* ######################################################### */
/*                         ScanMaxMin                        */
/* ######################################################### */
void  ScanMaxMin (frPtr, niPtr, fvPtr, maxD) 
        fileRepType    * frPtr;
        normInfoType   * niPtr ;
        FILE           * fvPtr  ;
        double           maxD ;
{
  uLong row, col ;
  double tempD ;

  for (row = 1; row < (frPtr->dataLines); row++)
  {  
     for (col = 0; col < (frPtr->embedDim); col++)
     { 
       {gFSCANFdBL("ScanMaxMin", frPtr->fileName, row+2, fvPtr, &tempD, maxD);}

       if (tempD > (niPtr->max)[col])  (niPtr->max)[col] = tempD ;
       if (tempD < (niPtr->min)[col])  (niPtr->min)[col] = tempD ;
     }
  }
   return ;
}
/* ######################################################### */
/*                         GetMaxDiff                        */
/* ######################################################### */
/*
     SIMPLE HELPER FUNCTION TO FIND THE LARGEST SPAN OF
     COORDINATE VALUES.  FOR EXAMPLE, IF POINTS HAVE TWO
     COORDINATES EACH, X AND Y, AND IF X RANGES BETWEEN 2 AND
     10, AND Y RANGES BETWEEN 13 AND 15, THEN THE FUNCTION
     BELOW CALCULATES THAT 10 - 2 = 8 IS THE maxDiff.

*/
void GetMaxDiff (normInfoPtr, maxDouble)
        normInfoType   * normInfoPtr ;
        double         maxDouble ;
{
  double temp ;
  int coord ;
  for (coord = 0; coord < normInfoPtr->embedDim; coord++)
  {
            /*
               THIS INDIRECTLY TESTS WHETHER MAX-MIN IS BIGGER
               THAN THE MAXDOUBLE.  IF SO, WE HAVE TO ABORT --
               THERE'S NO ROOM TO STORE MaxDiff IN A DOUBLE.
            */
     if ( ((normInfoPtr->max)[coord] >=0)
          && ((-(normInfoPtr->min)[coord]) > (maxDouble -
          (normInfoPtr->max)[coord]))   )
        MaxDiffOflowErr () ;
  }
  temp = (normInfoPtr->max)[0] - (normInfoPtr->min)[0] ;
  for (coord = 1; coord < normInfoPtr->embedDim; coord++)
  {
     if (temp < (normInfoPtr->max)[coord] -
                  (normInfoPtr->min)[coord] )
       temp = (normInfoPtr->max)[coord] -
                  (normInfoPtr->min)[coord] ;
  }
  (normInfoPtr->maxDiff) = temp ;
  return ;
}
/* ######################################################### */
/*                      MaxDiffOflowErr                      */
/* ######################################################### */
/*
    THIS NEEDS TO BE MADE A LITTLE MORE DIAGNOSTIC.
*/
void MaxDiffOflowErr () 
{
      printf("\n");
      printf("PROBLEM:\n");
      printf("\n");
      printf("Function \"GetMaxDiff\" failed due to an overflow.");
      printf("\n");
      printf("Sorry, fd3 must abort now.\n");
      printf("\n");
      fflush(stdout) ; exit(EXIT_FAILURE);
   }
/* ######################################################### */
/*                           NoScaleErr                      */
/* ######################################################### */
void NoScaleErr ()
   {
      printf("\n");
      printf("PROBLEM:\n");
      printf("\n");
      printf("Function \"GetMaxDiff\" is not able to create\n");
      printf("a scale for the data.  Check the data file for\n");
      printf("extreme values.\n");
      printf("\n");
      printf("Sorry, fd3 must abort now.\n");
      printf("\n");
      fflush(stdout) ; exit(EXIT_FAILURE);
   }
/* ######################################################### */
/*                      PrintNormztnInfo                     */
/* ######################################################### */
void PrintNormztnInfo (normInfoPtr)
        normInfoType   * normInfoPtr ;
{
   int coord ;
   printf ("\n" ) ;
   printf ("The NORMALIZATION INFORMATION is: \n" ) ;
   printf ("\n" ) ;
   printf ("ulongBitNum    = %d\n", normInfoPtr->ulongBitNum) ;
   printf ("embedDim       = %lu\n", normInfoPtr->embedDim) ;
   printf ("coord#            max            min\n\n") ;
   for (coord = 0; coord < normInfoPtr->embedDim; coord++)
   {printf ("%3d%18G%15G\n",coord,
         normInfoPtr->max[coord],normInfoPtr->min[coord]);}
   printf ("\n") ;
   printf ("maxDiff        = %1G\n", normInfoPtr->maxDiff) ;
   printf ("maxUlong       = %lu\n", normInfoPtr->maxUlong) ;
   printf ("scalingFactor  = %1G\n", normInfoPtr->scalingFactor) ;
   printf ("maxCoord       = %lu\n", normInfoPtr->maxCoord) ;
   printf ("\n") ;
   printf ("FD will transform your data as follows:\n") ;
   printf ("\n") ;
   printf ("FDpoint = (yourPoint - minvec) * scalingFactor\n") ;
   printf ("\n") ;
   printf ("\"Minvec\" is the vector of coordinate minimums -- the\n") ;
   printf ("third column of the table above.\n") ;
   printf ("\n") ;
   fflush(stdout);

   return ;
}
/* ######################################################### */
/*                          GetRAMsize                       */
/* ######################################################### */
void  GetRAMsize (machinePtr, fileRepPtr, ramDataPtr)
        machineType    * machinePtr;
        fileRepType    * fileRepPtr;
        ramDataRepType * ramDataPtr ;
{
  (ramDataPtr->ulongBitNum) = (machinePtr->ulongBitNum) ;
  (ramDataPtr->embedDim) =    (fileRepPtr->embedDim) ;
  (ramDataPtr->dataLines) =   (fileRepPtr->dataLines) ;

  return ;
}
/* ######################################################### */
/*                            GetRAM                         */
/* ######################################################### */
/*
    WE ALLOCATE, IN ESSENCE, A TWO-DIMENSIONAL ARRAY TO HOLD THE
    DATA.  EACH ROW IS ONE POINT.  FOR EXAMPLE, IF DATA POINTS
    HAVE TWO COORDINATES, WE ALLOCATE AN X-ARRAY (DESIGNATED
    pointList[0]) AND A Y-ARRAY (pointList[1]).  WE ALSO
    ALLOCATE STORAGE FOR A "nextAfter" ARRAY OF INTEGERS.
    THESE INTEGERS ARE USED AS INDEXES (POINTERS) TO IMPLEMENT
    NEW ORDERINGS OF DATA POINTS. 

    WE FIRST ALLOCATE pointList, AN ARRAY OF POINTERS.  THERE
    IS ONE POINTER TO POINT TO EACH COORDINATE ARRAY.  THEN WE
    ALLOCATE STORAGE FOR EACH COORDINATE ARRAY of the form
    pointList[coord].  FINALLY WE ALLOCATE STORAGE FOR A
    PARALLEL ARRAY OF POINTERS.

    WE DID NOT HAVE TO MODIFY THIS FUNCTION TO GET THE
    mincover OPTION WORKING.  HOWEVER, CHANGES MAY BE
    NECESSARY WHEN AND IF THE mincover OPTION IS UPGRADED TO
    HANDLE INPUTS WITH EMBEDDING DIMENSION GREATER THAN 1.
*/
void GetRAM (cP, rP)
        commandType    * cP;
        ramDataRepType * rP ;
{ 
  int coord ;
  size_t eD = (size_t) (rP->embedDim),
         dL = (size_t) (rP->dataLines),
         suls = (size_t) (sizeof(uLong *)) ,
         sul = (size_t) (sizeof(uLong)) ;
         
  if ( (cP->debugging) || (cP->checking) )
      {printf("GetRAM: Allocating storage ... ");fflush(stdout);}

  (rP->pointList) =  (uLong **)
     gmalloc ("GetRAM", "rP->pointList", eD * suls);

  for (coord = 0; coord < (eD); coord++)
     ((rP->pointList)[coord]) =
            (uLong *) gmalloc ("GetRAM", "rP->pointList", dL * sul) ;

  (rP->nextAfter) =
            (uLong *) gmalloc ( "GetRAM", "rP->nextAfter", dL * sul);

  if ( (cP->debugging) || (cP->checking) )
     {printf("Done.\n\n"); fflush(stdout);}

  return ; 
}
/* ######################################################### */
/*                           InitRAM                         */
/* ######################################################### */
/*
     THIS FUNCTION INITIALIZES THE QUEUES THAT ARE USED IN THE
     RADIX SORT.
    
     WE DID NOT HAVE TO MODIFY THIS FUNCTION TO GET THE
     mincover OPTION WORKING.  HOWEVER, CHANGES MAY BE
     NECESSARY WHEN AND IF THE mincover OPTION IS UPGRADED TO
     HANDLE INPUTS WITH EMBEDDING DIMENSION GREATER THAN 1.
*/
void InitRAM (ramDataPtr)
        ramDataRepType * ramDataPtr ;
{ 
  CreateQ (   &( (ramDataPtr->Q)[0] )   ) ;
  CreateQ (   &( (ramDataPtr->Q)[1] )   ) ;

  return ;
}
/* ######################################################### */
/*                           LoadRAM                         */
/* ######################################################### */
/*
    IN THIS FUNCTION WE READ IN EACH COORDINATE OF EACH DATA
    POINT.  WE NORMALIZE EACH COORDINATE BY SUBTRACTING THE
    MINIMIMUM ATTAINED BY THE DATA IN THAT COORDINATE POSITION,
    AND BY SCALING SO THE MAXIMUM VALUE ATTAINABLE BY ANY
    COORDINATE IS maxUlong, THE LARGEST VALUE THAT THE DATA
    TYPE "uLong" CAN HAVE.  FOR EXAMPLE, EACH X-COORDINATE IS
    NORMALIZED BY SUBTRACTING THE MINIMUM X-COORDINATE, AND
    THEN MULTIPLYING BY A SCALING FACTOR.  THE EFFECT IS TO FIT
    THE DATA "TIGHTLY" INTO A "BOUNDING BOX."  THE POINT IS
    THEN PLACED IN A QUEUE, TO AWAIT PROCESSING.  THE NEXT
    THING THAT WILL HAVE TO BE DONE TO THE DATA IS A RADIX
    SORT, WHICH IS THE REASON THAT IT IS BEING PUT A QUEUE
    HERE.  

    THE DATA-SCANNING LOOP NATURALLY RUNS IN TIME PROPORTIONATE
    TO THE INPUT SIZE (NUMBER OF POINTS IN THE DATA FILE).  TO
    SPEED THINGS UP WHEN THE DATA SET IS LARGE, WE USE MACRO'S
    "gFSCANFdBL" and "ENQ" INSTEAD OF FUNCTION CALLS.
    *WARNING* TO THOSE WHO MAKE CHANGES TO THIS CODE:  DON'T
    CHANGE *ANYTHING* IN MACRO CALLS OR DEFINITIONS, UNLESS YOU
    ARE QUITE SURE YOU KNOW WHAT YOU ARE DOING.

    WE DID NOT HAVE TO MODIFY THIS FUNCTION TO GET THE mincover
    OPTION WORKING.  HOWEVER, CHANGES MAY BE NECESSARY WHEN AND
    IF THE mincover OPTION IS UPGRADED TO HANDLE INPUTS WITH
    EMBEDDING DIMENSION GREATER THAN 1.
*/
void LoadRAM(cP, frP, rdP, niP, maxD)
       commandType    * cP   ;
       fileRepType    * frP  ;
       ramDataRepType * rdP  ;
       normInfoType   * niP  ;
       double           maxD ;
{ 
  FILE    * fp ;

  InitLR (cP, frP, &fp, maxD) ;

  LRScan ( (rdP->dataLines),     (rdP->embedDim),
           (rdP->pointList),     (rdP->nextAfter), niP,
           &((rdP->Q)[0]),       (frP->fileName), fp, maxD) ;
  
  FinLR (cP, rdP, &fp) ;

  return ;
}
/* ######################################################### */
/*                           InitLR                          */
/* ######################################################### */
/*
     GET THINGS STARTED FOR THE LOADING OF THE RAM.  WE PRINT A
     STATUS MESSAGE AND OPEN THE FILE.  LASTLY WE READ PAST THE
     FIRST NUMBER (dataLines) IN THE FILE, WHICH BRINGS US TO
     THE BEGINNING OF THE LIST OF DATA POINTS.
*/
void InitLR (cP, frP, fpP, maxD) 
       commandType    *   cP;
       fileRepType    *   frP;
       FILE           * * fpP ;
       double             maxD ;
{
   double iBf ;

   if   ( !(cP->terse))
   { printf("LoadRAM: Loading and normalizing data ... ") ; 
     fflush(stdout);   }
     
   Gfopen ("InitLR", fpP, (frP->fileName), "rb+") ;
   {gFSCANFdBL("InitLR", (frP->fileName), (uLong)1, (*fpP), (&iBf), maxD);}

   return ;
}
/* ######################################################### */
/*                           LRScan                          */
/* ######################################################### */
/*
      dLines = NUMBER OF LINES OF DATA.  eDim IS THE EMBEDDING
      DIMENSION. pList IS A POINTER TO AN ARRAY FOR STORING
      POINT VALUES, next IS AN ARRAY USED AS POINTERS TO MAKE
      ELEMENTS OF pList WORK AS A LINKED LIST, niP IS A POINTER
      TO INFORMATION NEEDED TO NORMALIZE (SCALE) THE POINT
      VALUES, Qp IS A POINTER TO AN ARRAY OF TWO QUEUES THAT
      ARE IMPLEMENTED USING THE LINKING APPARATUS OF pList AND
      next, fName IS THE NAME OF THE FILE THAT CONTAINS THE
      DATA POINTS, fp IS THE FILE VARIABLE FOR THE FILE, AND
      maxD IS THE VALUE OF THE LARGEST POSSIBLE QUANTITY OF
      DATA TYPE "double".

      THIS FUNCTION GOES THROUGH THE DATA FILE, READS IN EACH
      POINT, SCALES IT, PUTS IT INTO A ROW OF pList, AND THEN
      ENQUEUES IT.  THIS GETS THE SCALED DATA INTO MEMORY, AND
      SETS THINGS UP FOR A RADIX SORT.
*/
      
void LRScan(dLines,eDim,pList,next,niP,Qp,fName,fp, maxD) 
       uLong            dLines,      eDim,  **pList, *next ;
       normInfoType   * niP   ;
       QheaderType    * Qp    ;
       char           * fName ;
       FILE           * fp    ;
       double           maxD ;
{
    uLong ptNum,  coord ;
    double inBuf, normBuf, flrBuf, sFac, *minVect ;
    uLong maxUlong ;

   maxUlong = (niP->maxUlong) ;
   sFac = (niP->scalingFactor) ;
   minVect = (niP->min) ; /* vector of min coords */
   
   for(ptNum = 0; ptNum < dLines; ptNum++)
   { for (coord = 0; coord < eDim; coord++)
     {  {gFSCANFdBL("LRScan", fName, (uLong)ptNum+2, fp, &inBuf, maxD);}
        normBuf = (inBuf-minVect[coord]) * sFac ;
          /*
              MACHINE IMPRECISION *CAN* ALLOW THE SCALED VALUE
              TO GET OUTSIDE THE RANGE [0,maxUlong].  WE DO THE
              BEST WE CAN UNDER THE CIRCUMSTANCES -- "CLAMP" THE
              VALUE.
          */
        if   (normBuf >= maxUlong) pList[coord][ptNum] = maxUlong ;
        else if    (normBuf <= 0.0) pList[coord][ptNum] = 0 ;
             else    /* normBuf IS WITHIN (0,maxUlong) */
                 {
                /*
                     THE ROUNDING DONE HERE APPEARS TO ACCOUNT
                     FOR SMALL DIFFERENCES THAT ARE SOMETIMES
                     SEEN BETWEEN THE OUTPUT OF version 0.4 OF
                     FD3 AND PREVIOUS versions.  IN THE
                     PREVIOUS versions, CONVERSION WAS DONE
                     WITH A SIMPLE CAST, AND THAT CAST MAY
                     MERELY TRUNCATE.  SINCE WHAT WE ARE DOING
                     HERE IS MORE ACCURATE, WE WILL RETAIN IT,
                     EVEN THOUGH IT REDUCES "BACKWARD
                     COMPATABILITY" SLIGHTLY.
                */
                    flrBuf = floor(normBuf) ;
                    if   ( (normBuf - flrBuf) >= 0.5 )
                         pList[coord][ptNum] = ((uLong)flrBuf) + 1;
                    else pList[coord][ptNum] = ((uLong)flrBuf);
                 }
     }
     { ENQ(Qp, next, ptNum) ; }
   }
   return ;
}
/* ######################################################### */
/*                            FinLR                          */
/* ######################################################### */
void FinLR (cP, rdP, fpP )
       commandType    * cP    ;
       ramDataRepType * rdP   ;
       FILE           * * fpP ;
{
  if   ( !(cP->terse)){ printf("Done.\n") ; fflush(stdout) ;  }
  
  fclose( * fpP);

  if  ( (cP->debugging) && (cP->verbose) )
  {  printf ("AFTER COMPLETION OF \"LoadRAM\" ... \n\n") ;
     PrintRAMlists(rdP);  }

   return ;
}
/* ######################################################### */
/*                        PrintRAMlists                      */
/* ######################################################### */
/*
    DEVELOPER NOTES:

      HERE WE PRINT LISTS OF POINTS TO stdout.  DATA SETS CAN
      BE QUITE LARGE, SO THE OUTPUT OF printrRAMlists SHOULD GO
      TO A FILE.  CHOOSING THE FILE NAME IS A PROBLEM.
      IDEALLY, THE USER SHOULD BE ABLE TO SPECIFY A FILENAME.
      THERE COULD BE SOME DEFAULT NAME, BASED ON THE NAME OF
      THE INPUT FILE POSSIBLY, WHICH WOULD BE USED UNLESS THE
      FILE ALREADY EXISTS.  I DON'T WANT TO GET INTO MAKING ALL
      THAT WORK RIGHT NOW.  FOR NOW, LET'S LEAVE THIS ALONE.
      LATER, WE CAN ADD SOMETHING.  
*/
void PrintRAMlists (ramDataPtr)
       ramDataRepType   * ramDataPtr ;
{
  uLong     **pointList, * nextAfter, embedDim ;

  pointList = (ramDataPtr->pointList) ;
  nextAfter = (ramDataPtr->nextAfter) ;
  embedDim  = (ramDataPtr->embedDim)   ;

  printf ("QUEUE ZERO CONTAINS:\n\n") ;
  PrintQ( &((ramDataPtr->Q)[0]), nextAfter, pointList, embedDim ) ;

  printf ("QUEUE ONE CONTAINS:\n\n") ;
  PrintQ( &((ramDataPtr->Q)[1]), nextAfter, pointList, embedDim ) ;

  fflush(stdout);
  return ;
}
/* ######################################################### */
/*                           SortRAM                         */
/* ######################################################### */
/*
     THIS SORT IS TO BE USED DIRECTLY AFTER THE INITIAL LOADING
     OF THE DATA INTO Q[0].  IT LEAVES THE SORTED DATA IN Q[0].

     IN THE CASE OF EMBEDDING DIMENSION 1, THE SPECIAL SORT
     USED BY FD3 REDUCES TO A *STANDARD* SORT INTO ASCENDING
     ORDER.  FOR THIS REASON, WE DID NOT HAVE TO MODIFY THIS
     FUNCTION TO GET FD3's mincover OPTION WORKING.  HOWEVER,
     CHANGES MAY BE NECESSARY WHEN AND IF THE mincover OPTION
     IS UPGRADED TO HANDLE INPUTS WITH EMBEDDING DIMENSION
     GREATER THAN 1.

*/
void  SortRAM(cmdP, rDptr)
        commandType    * cmdP;
        ramDataRepType * rDptr ;
{
  int   bitpos, ulongBitNum;
  uLong offset, embedDim ;

  embedDim    =  (rDptr->embedDim) ;
  ulongBitNum = (rDptr->ulongBitNum) ;

  SRstartMsg (cmdP) ;

  for (bitpos = 0; bitpos < ulongBitNum; bitpos++)
      for (offset = 1; offset <= embedDim ; offset++)
          {rADpASS(rDptr, embedDim - offset, bitpos);}

  ConcatQs (&((rDptr->Q)[0]), &((rDptr->Q)[1]), rDptr->nextAfter) ;

  SRfinMsg (cmdP, rDptr) ;

  return ;
}
/* ######################################################### */
/*                        SRstartMsg                         */
/* ######################################################### */
void  SRstartMsg (commandPtr) 
        commandType    * commandPtr;
{
  if   ( !(commandPtr->terse))
  {
     printf("SortRAM: Sorting data ... ");
     fflush(stdout);
  }

  return ;
}
/* ######################################################### */
/*                           RadPass                         */
/* ######################################################### */
/*
     THIS PROCEDURE PERFORMS ONE PASS OF THE RADIX SORT.
*/
void RadPass(rDptr, coord, bitpos)
   ramDataRepType  *rDptr ;
   uLong            coord;   /* COORDINATE WE ARE DOING THIS PASS ON */
   int              bitpos;  /* BIT POSITION WE ARE DOING THIS PASS ON */
{
   int sQnum ;
   uLong iSize[2], rmnder ;

   for (sQnum = 0; sQnum < 2; sQnum++ )
        iSize[sQnum]= SizeQ (&((rDptr->Q)[sQnum])) ;

   for (sQnum = 0; sQnum < 2; sQnum++)
   {
     if (iSize[sQnum] != 0)
     {
       {REqSRcRUN(rDptr,coord,bitpos,(uLong)sQnum,iSize[sQnum],&rmnder);}
       if (rmnder > 0)
         {REqGENcASE(rDptr,coord,bitpos,(uLong)sQnum, rmnder)};
     }
   }
   return ;
} 
/* ######################################################### */
/*                         ReQsrcRun                         */
/* ######################################################### */
/*
    THIS IS A HELPER FUNCTION FOR A RADIX SORT.  WE AIM TO
    ACHIEVE HIGH SPEED HERE.  THIS FUNCTION WILL BE RE-WRITTEN
    AS A MACRO.  THE FUNCTION REQUEUE'S THAT INITIAL RUN OF
    ELEMENTS IN THE SOURCE QUEUE WHICH HAVE TO COME OFF THE
    FRONT AND GO BACK ON TO THE REAR OF THE SOURCE QUEUE.  OF
    COURSE, THIS RUN MAY HAVE LENGTH ZERO!  THIS FUNCTION
    "CLEANS UP" AFTER ITSELF BY REQUEUEING THE LAST ELEMENT IT
    DEQUEUES.  THIS *LAST* ELEMENT MAY GO *NOT* INTO THE SOURCE
    QUEUE, BUT INTO THE *OTHER* QUEUE.

    PRECONDITIONS: PARAMETER "reQable" IS THE MAXIMUM NUMBER OF
    ELEMENTS IN THE QUEUE THAT THIS FUNCTION CALL IS ALLOWED TO
    REQUEUE.  "reQable" IS POSITIVE. (IN PARTICULAR, THE SOURCE
    QUEUE IS NOT EMPTY.)

    POSTCONDITION: EITHER THIS FUNCTION DEQUEUES ALL "reQable"
    ELEMENTS FROM THE SOURCE QUEUE, OR IT ENQUEUES AN ELEMENT
    ONTO THE *OTHER* QUEUE, OR BOTH.  
*/
void ReQsrcRun (ramDataPtr, coord, bitpos, srcQnum, reQable, remgPtr)
   ramDataRepType    *ramDataPtr ;
   uLong coord   ;  /* COORDINATE WE ARE DOING THIS PASS ON */
   int   bitpos  ;  /* BIT POSITION WE ARE DOING THIS PASS ON */
   uLong srcQnum ;  /* INDEX OF THE SOURCE QUEUE  */
   uLong reQable, * remgPtr ; 
{   uLong eltIndex, * indexArray, ** ptList ;
    QheaderType  *QsPtr ;
    indexArray = (ramDataPtr->nextAfter) ;
    ptList = (ramDataPtr->pointList) ;
    QsPtr = (ramDataPtr->Q) ;
    {DEQ(&(QsPtr[srcQnum]), indexArray, &eltIndex);}
    reQable-- ;
    while((reQable > 0) && ISa(srcQnum,ptList[coord][eltIndex],bitpos)) 
    { {ENQfAST(&(QsPtr[srcQnum]), indexArray, eltIndex);}
      {DEQ(&(QsPtr[srcQnum]), indexArray, &eltIndex);}
      reQable-- ;    }
    if ISa((uLong) 0,ptList[coord][eltIndex],bitpos)
          {ENQ(&(QsPtr[0]),indexArray,eltIndex);}
    else  {ENQ(&(QsPtr[1]),indexArray,eltIndex);}
    *remgPtr = reQable ;
    return ;
}    
/* ######################################################### */
/*                         ReQgenCase                        */
/* ######################################################### */
/*
    REQUEUE FROM THE SOURCE QUEUE, WITH THE PRECONDITION THAT
    NEITHER QUEUE IS EMPTY AND THERE ARE NO LESS THAN "reQable"
    ELEMENTS IN THE SOURCE QUEUE.  THIS FUNCTION TAKES "reQable"
    ELEMENTS OUT OF THE SOURCE QUEUE AND REQUEUES THEM USING
    THE FAST ENQUEUE OPERATION FOR ALL BUT THE LAST ELEMENT.
    UNDER THESE CONDITIONS, NO ATTEMPT TO ADD AN ELEMENT TO AN
    EMPTY QUEUE CAN OCCUR.
*/
void ReQgenCase (ramDataPtr, coord, bitpos, srcQnum, reQable)
   ramDataRepType    *ramDataPtr ;
   uLong coord ;    /* COORDINATE WE ARE DOING THIS PASS ON.   */
   int  bitpos ;    /* BIT POSITION WE ARE DOING THIS PASS ON. */
   uLong srcQnum ;  /* INDEX OF THE SOURCE QUEUE.  */
   uLong reQable ; 
{  uLong eltNum, eltIndex, * indexArray, ** ptList ;
   QheaderType  *QsPtr ;
   indexArray = (ramDataPtr->nextAfter) ;
   ptList = (ramDataPtr->pointList) ;
   QsPtr = (ramDataPtr->Q) ;
   for (eltNum = 0; eltNum < reQable - 1; eltNum++)
   { {DEQfAST(&(QsPtr[srcQnum]), indexArray, &eltIndex);}
     if ISa((uLong) 0,ptList[coord][eltIndex],bitpos)
           {ENQfAST(&(QsPtr[0]),indexArray,eltIndex);}
     else  {ENQfAST(&(QsPtr[1]),indexArray,eltIndex);}   }
   {DEQ(&(QsPtr[srcQnum]), indexArray, &eltIndex);}
   if ISa((uLong) 0,ptList[coord][eltIndex],bitpos)
         {ENQ(&(QsPtr[0]),indexArray,eltIndex);}
   else  {ENQ(&(QsPtr[1]),indexArray,eltIndex);}
   return;
}
/* ######################################################### */
/*                         SRfinMsg                          */
/* ######################################################### */
void  SRfinMsg (cmdP, rDptr) 
        commandType    * cmdP;
        ramDataRepType * rDptr ;
{
  if   ( !(cmdP->terse)) { printf("Done.\n"); fflush(stdout);  }

  if  ( (cmdP->debugging) && (cmdP->verbose) )
  {
     printf ("AFTER COMPLETION OF \"SortRAM\" ... \n\n") ;
     PrintRAMlists(rDptr);
     PrintRAMmesh (rDptr) ;
  }

   return ;
}
/* ######################################################### */
/*                        PrintRAMmesh                       */
/* ######################################################### */
void PrintRAMmesh (ramDataPtr)
       ramDataRepType   * ramDataPtr ;
{
  uLong     **pointList, * nextAfter, embedDim ;
  int ulongBitNum ;

  pointList   = (ramDataPtr->pointList)   ;
  nextAfter   = (ramDataPtr->nextAfter)   ;
  embedDim    = (ramDataPtr->embedDim)    ;
  ulongBitNum = (ramDataPtr->ulongBitNum) ;

  printf ("QUEUE ZERO MESH IS:\n\n") ;
  PrintQmesh( &((ramDataPtr->Q)[0]), nextAfter,
                   pointList, embedDim, ulongBitNum ) ;

  printf ("QUEUE ONE MESH IS:\n\n") ;
  PrintQmesh( &((ramDataPtr->Q)[1]), nextAfter,
                   pointList, embedDim, ulongBitNum ) ;

  fflush(stdout);
  return ;
}
/* ######################################################### */
/* ######################################################### */
/* ######################################################### */     
/*                     end datarep.c                         */
/* ######################################################### */
/* ######################################################### */
/* ######################################################### */
