
/* BEGIN COPYRIGHT NOTICE

Copyright (c) 1992 by John Sarraille and Peter DiFalco
(john@ishi.csustan.edu)

Copyright (c) 1998 by John Sarraille
(john@ishi.csustan.edu)

Permission to use, copy, modify, and distribute this software
and its documentation for any purpose and without fee is hereby
granted, provided that the above copyright notice appear in all
copies and that both that copyright notice and this permission
notice appear in supporting documentation.

The algorithm used in this program was inspired by the paper
entitled "A Fast Algorithm To Determine Fractal Dimensions By
Box Counting", which was written by Liebovitch and Toth, and
which appeared in the journal "Physics Letters A", volume 141,
pp 386-390, (1989).

This program is not warranted: use at your own risk.

END COPYRIGHT NOTICE */


/*

   FD3 PROGRAM FILE

Filename:  estim.c

OUTLINE OF CALLS:

InitRefEstD +
MakeRefFits +
  FitLSqrLine +
    Quotient +
InitEstData +
MakeFits +
  FitLSqrLine +
ReportDimEsts +
  PrintNoEstMsg +
  PrintRange +
  PrintDimEsts +
*/

#include "fd.h"

/* 
    THE FUNCTIONS IN THIS FILE: 
*/
void    InitRefEstD   () ; 
void    MakeRefFits   () ;
void    InitEstData   () ;
void    MakeFits      () ;
double  Quotient      () ;
void    FitLSqrLine   () ;
void    ReportDimEsts () ;
void    PrintNoEstMsg () ;
void    PrintRange    () ;
void    PrintDimEsts  () ;

/*   
      THE FUNCTIONS NOT IN THIS FILE,
      BUT CALLED FROM HERE:
*/
/* extern void * gmalloc    () ;  in mem.c */

/* ######################################################### */
/*                       InitRefEstD                         */
/* ######################################################### */
/*

*/
void  InitRefEstD (deP, rfcdP) 
        dimEstimDataType  * deP ;    refinedCDtype * rfcdP ;
{
   (deP->fineMeshPoint)   = (rfcdP->fineMeshPoint) ;
   (deP->saturationPoint) = (rfcdP->saturationPoint) ;

   return ;
}
/* ######################################################### */
/*                       MakeRefFits                         */
/* ######################################################### */
/*
    THIS PROCEDURE IS EXACTLY LIKE MakeFits BELOW, EXCEPT THAT
    IT TAKES rfcdP AS A PARAMETER INSTEAD OF cDptr.  THIS IS
    NECESSARY BECAUSE A DIFFERENT DATA TYPE IS USED TO HOLD THE
    COVERING STATISTICS USED IN A REFINED DIMENSION ESTIMATE.
    MAYBE THAT SHOULD CHANGE IN THE *NEXT* VERSION OF THIS
    SOFTWARE, IF THERE EVER IS ONE.
*/
void MakeRefFits (deP, rfcdP)
       dimEstimDataType * deP ;   refinedCDtype * rfcdP ;
{
   double  slope, intercept, RSqr;
   uLong first, last ;

   first = deP->saturationPoint ;
   last  = deP->fineMeshPoint ;

   if (first < last)
   {
      FitLSqrLine ( (rfcdP->logCellSize), (rfcdP->negLogCellCount),
                     first, last, &slope,  &intercept, &RSqr) ;
      (deP->capDimEst)     = slope ;
      (deP->capIntercept)  = intercept ;
      (deP->capRSqr)     = RSqr ;
   
      FitLSqrLine ( (rfcdP->logCellSize), (rfcdP->information),
                     first, last, &slope,  &intercept, &RSqr) ;
      (deP->infDimEst)     = slope ;
      (deP->infIntercept)  = intercept ;
      (deP->infRSqr)     = RSqr ;
   
      FitLSqrLine ( (rfcdP->logCellSize), (rfcdP->logSumSqrFreq),
                     first, last, &slope,  &intercept, &RSqr) ;
      (deP->corrDimEst)    = slope ;
      (deP->corrIntercept) = intercept ;
      (deP->corrRSqr)    = RSqr ;
   }

   return ;
}
/* ######################################################### */
/*                       InitEstData                         */
/* ######################################################### */
/*
        WE MERELY TRANSFER THE CUTOFF INDICES FROM ONE RECORD
        TO ANOTHER.  WE'D LIKE TO ADD SOME ABILITY FOR THE USER
        TO MODIFY THESE VALUES AS DESIRED.  ALSO, IF WE ADD A
        "SECOND PASS" DIMENSION ESTIMATE, FD3 SHOULD BE ABLE TO
        WIDEN THE RANGE OF STATISTICS USABLE IN THE DIMENSION
        ESTIMATES.  MOST USERS ARE WORKING WITH SMALL DATA
        SETS, AND THE WIDENING COULD BENEFIT THEM QUITE A LOT.
*/
void  InitEstData (cDptr, dEptr) 
      coverDataType     *  cDptr  ;
      dimEstimDataType  *  dEptr  ;
{
   (dEptr->fineMeshPoint) = (cDptr->fineMeshPoint) ;
   (dEptr->saturationPoint) = (cDptr->saturationPoint) ;

   return ;
}
/* ######################################################### */
/*                          MakeFits                         */
/* ######################################################### */
/*
    THE PURPOSE OF THIS FUNCTION IS TO INVOKE THE PROCEDURES
    THAT ESTIMATE THE VARIOUS VERSIONS OF FRACTAL DIMENSION.
    ALSO, IT RECORDS THE RESULT OF EACH ESTIMATE.  AT THIS
    TIME, THE ONLY PROCEDURE IT NEEDS TO CALL IS ONE THAT FITS
    A LEAST SQUARE LINE TO A SEQUENCE OF POINTS. THIS FUNCTION
    DOES ESSENTIALLY *NOTHING* IF THE CALCULATED SATURATION
    POINT OF THE COVERING STATISTICS IS NOT LESS THAN THE FINE
    MESH POINT.  IN PARTICULAR, IT LEAVES THE STORAGE FOR THE
    ESTIMATES UNTOUCHED IN THAT CASE.
*/
void  MakeFits (cDptr, dEptr)
      coverDataType     *  cDptr  ;
      dimEstimDataType  *  dEptr  ;
{
   double  slope, intercept, RSqr;
   uLong first, last ;

   first = dEptr->saturationPoint ;
   last  = dEptr->fineMeshPoint ;

   if (first < last)
   {
      FitLSqrLine ( (cDptr->logCellSize), (cDptr->negLogCellCount),
                     first, last, &slope,  &intercept, &RSqr) ;
      (dEptr->capDimEst)     = slope ;
      (dEptr->capIntercept)  = intercept ;
      (dEptr->capRSqr)     = RSqr ;
   
      FitLSqrLine ( (cDptr->logCellSize), (cDptr->information),
                     first, last, &slope,  &intercept, &RSqr) ;
      (dEptr->infDimEst)     = slope ;
      (dEptr->infIntercept)  = intercept ;
      (dEptr->infRSqr)     = RSqr ;
   
      FitLSqrLine ( (cDptr->logCellSize), (cDptr->logSumSqrFreq),
                     first, last, &slope,  &intercept, &RSqr) ;
      (dEptr->corrDimEst)    = slope ;
      (dEptr->corrIntercept) = intercept ;
      (dEptr->corrRSqr)    = RSqr ;
   }

   return ;
}
/* ######################################################### */
/*                        Quotient                           */
/* ######################################################### */
/*
     THIS IS JUST A LITTLE FUNCTION TO DO DIVISION BY
     SUCCESSIVE ADDITION IN A WAY THAT HANDLES THE "DIVISION BY
     ZERO" PROBLEM.

     IF SUCCESSFUL, THIS FUNCTION RETURNS AN ESTIMATE OF THE
     QUOTIENT OF THE FIRST TWO PARAMETERS (numer/denom).  THE
     VALUE RETURNED IS OF THE FORM (M*tolerance) WHERE M IS A
     NON-NEGATIVE INTEGER.

     PRECONDITIONS: ALL THE PARAMETERS ARE NON-NEGATIVE.
     
     IF THIS FUNCTION IS SUCCESSFUL THE ESTIMATE, quotEst, THAT
     IT RETURNS HAS THE PROPERTY THAT 

     (numer/denom) <= quotEst < (numer/denom) + tolerance.

     FAILURE IS DECLARED IF SUCCESSIVE ADDITIONS FAIL TO
     PRODUCE A VALUE OF quotEst CONFORMING TO THE INEQUALITIES
     ABOVE *BEFORE* quotEst ATTAINS A VALUE GREATER THAN THE
     PARAMETER "limit".

     WHEN FAILURE OCCURS, A VALUE OF -1 IS RETURNED.

     THE limit PARAMETER IS SUPPOSED TO BE AN UPPER BOUND ON
     WHAT THE VALUE OF (numer/denom) COULD POSSIBLY BE.
     FAILURE MAY MEAN THAT THE VALUE OF denom IS UNUSUALLY
     CLOSE TO ZERO, OR THAT THE VALUE OF limit IS NOT REALISTIC.
*/

double Quotient (numer, denom, tolerance, limit)
     double   numer, denom, tolerance, limit ;
{
  double quotEst = 0, numEst = 0, qIncr = tolerance,
  nIncr = tolerance * denom ;

  while ((numEst < numer) && (quotEst < limit))
  {
    quotEst += qIncr ;
    numEst  += nIncr ;
  }
  
 if (numEst >= numer) return (quotEst) ;
 else return (-1) ; /* failure */
}

/* ######################################################### */
/*                      FitLSqrLine                          */
/* ######################################################### */
/*
    FIT A LEAST SQUARE LINE TO DATA IN X,Y AND REPORT THE SLOPE
    AND INTERCEPT OF THE LINE.  NO PROTECTION IS PROVIDED HERE
    AGAINST OVERFLOW.  IT IS ASSUMED THAT last > first AND THAT
    THE X'S ARE NOT ALL THE SAME -- ELSE DIVISION BY ZERO WILL
    OCCUR.
*/

void FitLSqrLine (X, Y, first, last, slopePtr,
                        interceptPtr, RSqrErrPtr)
     double   X[], Y[] ;
     uLong    first, last ;
     double   *slopePtr, *interceptPtr, *RSqrErrPtr ;
{
  uLong  indx, pointCount ;
  double Xsum=0, Ysum=0, XYsum=0, XXsum=0, YYsum=0,
         Xmean=0, Ymean=0, Xtemp, Ytemp, SSdueToRegr,
         SSaboutMean ;

  for (indx = first; indx <= last; indx++)
  {
     Xtemp = X[indx];
     Ytemp = Y[indx];
     Xsum += Xtemp;
     Ysum += Ytemp;
     XYsum += (Xtemp * Ytemp);
     XXsum += (Xtemp * Xtemp);
     YYsum += (Ytemp * Ytemp);
  }
  pointCount = last - first + 1 ;
  Xmean = Xsum/pointCount;
  Ymean = Ysum/pointCount;
  *slopePtr = (XYsum - (Xsum * Ymean))/(XXsum - (Xsum * Xmean)) ;
  *interceptPtr = Ymean - (*slopePtr * Xmean) ;

  SSdueToRegr = (*slopePtr)*(*slopePtr)*((XXsum)-(Xsum*Xmean)) ;
  SSdueToRegr = fabs(SSdueToRegr) ;
  
  SSaboutMean = (YYsum)-(Ysum*Ymean) ;
  SSaboutMean = fabs(SSaboutMean) ;

  /* HERE WE MAKE A STAB AT AVOIDING PROBLEMS WITH VALUES OF
     THE SUMS OF SQUARES THAT ARE NEAR ZERO. */

  /* IF SSaboutMean IS ZERO, THEN SSdueToRegr SHOULD BE ZERO,
     AND THE FIT SHOULD BE PERFECT, SO WE WILL JUST ASSIGN A
     VALUE OF 1 TO THE R-square NUMBER */
     
  if ((SSaboutMean == 0))  *RSqrErrPtr = 1 ;
  else
  {
    *RSqrErrPtr = Quotient(SSdueToRegr, SSaboutMean,
                   (double)(1.00/10000), (double)(1.00) ) ;
      /*
         IF THE FAILURE VALUE IS RETURNED BY QUOTIENT, THEN IT
         PROBABLY MEANS SSaboutMean IS VERY CLOSE TO ZERO.  IN
         THAT CASE, THE FIT IS PROBABLY VERY GOOD -- JUST A LINE
         WITH SLOPE EQUAL TO ZERO.  SO IN THAT CASE, WE'LL JUST
         CHANGE THE VALUE TO 1, INDICATING A PERFECT FIT.
      */
    if ( (*RSqrErrPtr < 0) || (*RSqrErrPtr > 1))
    {
      *RSqrErrPtr = 1 ;
    }
  }
  return ;
}
/* ######################################################### */
/*                   ReportDimEsts                           */
/* ######################################################### */
void ReportDimEsts (dEptr) 
       dimEstimDataType  * dEptr ;
{
   if ( (dEptr->saturationPoint) >= (dEptr->fineMeshPoint) )
      PrintNoEstMsg ();
   else
   {
      PrintRange ();
      PrintDimEsts  (dEptr);
   }
   return ;
}
/* ######################################################### */
/*                      PrintNoEstMsg                        */
/* ######################################################### */
void PrintNoEstMsg ()
{
   printf("\n") ;

   printf("Fd3 is unable to make its \"automatic\" dimension.\n") ;
   printf("estimates using this data set.  Fd3 failed to find\n") ;
   printf("a suitable cell-size range.  Covering statistics\n") ;
   printf("reach saturation too quickly.\n") ;
   printf("\n") ;
   printf("If you wish, you can run the program again, using the\n") ;
   printf("-p option to specify a cell-size range.  Fd3 will then\n") ;
   printf("give you dimension estimates based on that range.\n") ;
   printf("\n") ;

   return ;
}
/* ######################################################### */
/*                       PrintRange                          */
/* ######################################################### */
void  PrintRange ()
{
   printf ("\n\n\n") ;
   printf ("The \"star bars\" in the lists above indicate the range\n") ;
   printf ("of covering statistics which was used to compute the\n") ;
   printf ("overall dimension estimates below.\n") ;
   printf("\n") ;
   return ;
}
/* ######################################################### */
/*                   PrintDimEsts                            */
/* ######################################################### */
void PrintDimEsts (dEptr)
       dimEstimDataType  * dEptr ;
{
   printf("\n\n") ;
   printf("Least-Square Estimates based on Indicated Cell Range:");
   printf("\n\n") ;

   printf("Fractal Dimension  (Capacity)   =  " ) ;
   printf("%.5f", (dEptr->capDimEst) );
   printf("     (R-sqr = ") ;
   printf("%.2f%%)\n", 100*(dEptr->capRSqr) );

   printf("Fractal Dimension (Information) =  ");
   printf("%.5f", (dEptr->infDimEst) );
   printf("     (R-sqr = ") ;
   printf("%.2f%%)\n", 100*(dEptr->infRSqr) );

   printf("Fractal Dimension (Correlation) =  ");
   printf("%.5f", (dEptr->corrDimEst));
   printf("     (R-sqr = ") ;
   printf("%.2f%%)\n", 100*(dEptr->corrRSqr) );

   printf("\n\n****************************************\n");    

   return ;
}
/* ######################################################### */
/* ######################################################### */
/* ######################################################### */     
/*                       end estim.c                         */
/* ######################################################### */
/* ######################################################### */
/* ######################################################### */

