
/* BEGIN COPYRIGHT NOTICE

Copyright (c) 1992 by John Sarraille and Peter DiFalco
(john@ishi.csustan.edu)

Copyright (c) 1998 by John Sarraille
(john@ishi.csustan.edu)

Permission to use, copy, modify, and distribute this software
and its documentation for any purpose and without fee is hereby
granted, provided that the above copyright notice appear in all
copies and that both that copyright notice and this permission
notice appear in supporting documentation.

The algorithm used in this program was inspired by the paper
entitled "A Fast Algorithm To Determine Fractal Dimensions By
Box Counting", which was written by Liebovitch and Toth, and
which appeared in the journal "Physics Letters A", volume 141,
pp 386-390, (1989).

This program is not warranted: use at your own risk.

END COPYRIGHT NOTICE */

/*
     HEADER FILE FOR FD3 SOFTWARE
     
     Filename:  fd.h
*/

/* ########################################################## */
/*                     INCLUDE FILES                          */
/* ########################################################## */

#include <stdio.h>
#include <ctype.h>
#include <errno.h>
/* #include <sys/ieeefp.h> */
/* #include <floatingpoint.h> */
#include <float.h>
#include <limits.h>
#include <math.h>
/* #include <stddef.h> */
#include <stdlib.h>
#include <string.h>

/* ########################################################## */
/*                         MACROS                             */
/* ########################################################## */

#ifndef FALSE
#define FALSE                 (0)
#endif

#ifndef TRUE
#define TRUE                  (1)
#endif

#ifdef   FILENAME_MAX
#define  MAXfNAMElen FILENAME_MAX
#endif

#ifndef FILENAME_MAX
#define MAXfNAMElen         (255)
#endif

#ifndef DBL_MAX
#define DBL_MAX      (1.797693e+308)
#endif

#ifndef EXIT_FAILURE
#define EXIT_FAILURE      (1)
#endif

#ifndef SEEK_CUR
#define SEEK_CUR      (1)
#endif

/*
    MACRO TO CALCULATE THE n-PLACE LEFT SHIFT OF x.  IT IS
    INTENDED THAT x BE AN unsigned long int, AND THAT n BE SOME
    NON-NEGATIVE INTEGER NO GREATER THAN ulongBitNum - 1
    (DEFINED BELOW).  A CAST IS DONE TO x "JUST IN CASE".  THE
    VALUE OF THE EXPRESSION WILL BE unsigned long int.
*/
#define LsHIFT(x,n) (((unsigned long int) (x)) <<  (n))


/*
    MACRO TO RETURN TRUE IF THE nTH BIT OF x IS A y, ELSE
    FALSE.  HERE x AND n SHOULD BE AS FOR LsHIFT, AND y SHOULD
    BE AN unsigned long int EQUAL TO 1 OR 0.
*/
#define ISa(y,x,n) (((x) & LsHIFT(1,n)) == (LsHIFT(y,n)))



/* ########################################################## */
/*                   TYPE DEFINITIONS                         */
/* ########################################################## */


/* ########################################################## */
/*                         uLong                              */
/* ########################################################## */
/*
      ALIAS FOR unsigned long int
*/

typedef unsigned long int uLong ;

/* ########################################################## */
/*                      commandType                           */
/* ########################################################## */
/*
      A STRUCTURE TO REPRESENT THE COMMAND INPUT BY THE USER
*/

  typedef struct
    {
      char  infileName[MAXfNAMElen + 1] ;
        /*
           FOR ERROR CHECKING
        */
      int   infileNameLength ;

        /*
           OPTION FLAGS
        */
      int checking   ;
      int debugging  ;
          /*
             CREATE A FINE MESH: INTERPOLATE MORE CELL SIZES IN
             BETWEEN THE SIZES USED BY THE STANDARD ALGORITHM.
             FOR EXAMPLE, INSTEAD OF CELL SIZES OF 1, 2, 4, 8,
             ...  WE COULD HAVE SIZES OF 1, 3/2, 2, 3, 4, 6, 8,
             ...  (CURRENTLY THIS OPTION CAN ONLY BE USED TO
             GET 4 TIMES THE DEFAULT NUMBER OF CELL SIZES.)
          */
      int fine       ;
          /*
             MINCOVER: DO THE ESTIMATION A SPECIAL WAY: USE AN
             ALGORITHM THAT CALCULATES THE EXACT CARDINALITY OF
             MINIMAL COVERINGS OF THE DATA POINTS.
          */
      int mincover   ;
          /*
             SCALING INFORMATION: TELL THE USER WHAT MAPPING
             WAS USED TO TRANSFORM HIS DATA INTO THE COORDINATE
             SYSTEM USED BY FD. 
          */
      int scaleInfo  ;
      int terse      ;
      int verbose    ;
        /*
            WITH THIS OPTION, THE USER CAN ASK THE THAT THE
            "STARBARS" BE PLACED AS HE LIKES.  THIS ALLOWS THE
            USER TO GET THE PROGRAM TO MAKE THE DIMENSION
            ESTIMATE BASED ON A RANGE OF CELL SIZES CHOSEN BY
            THE USER.
        */
      int placeBars  ;
        /*
            NUMERICAL PARAMETERS
        */

         /*
            MAGNIFIER DESCRIBES THE PROGRESSION OF CELL SIZES.
            EACH NEXT-LARGEST CELL SIZE IS 'MAGNIFIER' TIMES
            LARGER THAN THE PREVIOUS CELL SIZE.
         */
       double magnifier ;

       uLong    numCellSizes  ;
         /*
             INDEXES OF THE SMALLEST AND LARGEST CELL SIZES TO
             BE USED IN CALCULATING THE DIMENSION ESTIMATES.
         */
       uLong    saturationPoint ;
       uLong    fineMeshPoint   ;
        /*
              ERROR FLAGS
        */
      int invalidOption ;
      int repeatedOption ;
          /*
              WE CAN'T ALLOW SOME OPTIONS TO BE USED TOGETHER.
          */
      int mutexOptions ;
               
    } commandType ;

/* ########################################################## */
/*                      machineType                           */
/* ########################################################## */
/*
      A STRUCTURE TO REPRESENT CHARACTERISTICS OF THE VIRTUAL
      MACHINE ON WHICH THE PROGRAM RUNS.  AT THIS TIME, WE ONLY
      RECORD THE NUMBER OF BYTES AND BITS USED TO REPRESENT
      CERTAIN DATA TYPES, ALONG WITH THE MAXIMUM VALUES OF SOME
      DATA TYPES.
*/      

  typedef struct
    {
      int intByteNum ;
      int intBitNum  ;
      int maxInt ;
      
      int longByteNum ;
      int longBitNum  ;
      
      int ulongByteNum ;
      int ulongBitNum ;

      uLong  maxUlong ;

      int doubleByteNum ;
      int doubleBitNum  ;
      double maxDouble ;
      
    } machineType ;

/* ########################################################## */
/*                      fileRepType                           */
/* ########################################################## */
/*
      A STRUCTURE FOR REPRESENTING AN FD3 DATA FILE
*/      

typedef struct
{
   char                 fileName [MAXfNAMElen+1] ;
   FILE               * fileVarPtr ;
          /*
             INFO ABOUT THE DIMENSIONS OF THE ARRAY OF DATA
             STORED IN THE FILE
              columns     rows
              -------     ----
          */
   uLong   embedDim, dataLines ;

} fileRepType ;

/* ########################################################## */
/*                      normInfoType                          */
/* ########################################################## */
/*
      A STRUCTURE FOR STORING THE INFORMATION NEEDED TO
      RE-NORMALIZE (SHIFT AND SCALE) FD3 DATA SO THAT ALL
      COORDINATES ARE EXPRESSED AS UNSIGNED LONG INTEGERS WITH
      VALUES BETWEEN 0 AND maxUlong (INCLUSIVE).
*/      
  typedef struct
    {
          /*
             INFO ABOUT THE SIZE OF THE ATOMIC TYPES USED HERE.
          */
      int               ulongBitNum ;
          /*
             SIZE OF THE ARRAYS STORED  HERE.
         */
      uLong     embedDim ;
         /*
             ARRAYS OF MINIMUMS AND MAXIMUMS.  USED FOR STORING
             THE MIN AND MAX ATTAINED BY EACH COORDINATE OF THE
             DATA STORED IN SOME FD3 FILE.
         */
      double *          min ;
      double *          max ;
         /*
             THE MAXIMUM, TAKEN OVER THE DIFFERENCES BETWEEN THE
             MAXIMUM AND MINIMUM OF RESPECTIVE COORDINATE VALUES
             OF THE GIVEN SET OF DATA POINTS.  MaxDiff IS THUS
             THE SIDE LENGTH OF THE SMALLEST POSSIBLE "STANDARD
             CELL" (BOX) LARGE ENOUGH TO CONTAIN THE ENTIRE SET
             OF DATA POINTS CORRESPONDING TO THE min AND max
             ARRAYS.  (A "STANDARD CELL" IS ORIENTED WITH ITS
             SIDES PARALLEL TO THE COORDINATE PLANES.)
         */
      double            maxDiff ;
         /*
             THE LARGEST UNSIGNED INTEGER THAT CAN BE EXPRESSED
             (ASSUMING STANDARDIZED PRACTICE OF C COMPILERS)
             WHEN THE NUMBER OF BITS IN THE INTEGER IS
             ulongBitNum.  maxUlong WILL BE THE MAXIMUM OF THE
             SET OF NORMALIZED DATA COORDINATES.
         */
      uLong maxUlong ;
         /*
             THE SCALING FACTOR WHICH IS USED IN THE
             RENORMALIZATION.  THIS EXPRESSES HOW MUCH THE SET
             IS EXPANDED OR CONTRACTED BY THE RENORMALIZATION.
             SINCE A SINGLE SCALING FACTOR IS APPLIED TO ALL
             COORDINATES, THERE IS NO DISTORTION OF THE DATA.
         */
      double            scalingFactor ;
         /*
             THIS IS THE LARGEST VALUE THAT WILL BE FOUND AMONG
             THE COORDINATES OF THE NORMALIZED SET OF DATA
             POINTS.  IN "THEORY", THE VALUE OF maxCoord SHOULD
             BE THE SAME AS THE LARGEST POSSIBLE VALUE OF AN
             UNSIGNED LONG INT.  HOWEVER, THERE IS SOME
             INSTABILITY INHERENT IN THE WAY THAT THE SCALING
             FACTOR IS CALCULATED.  THIS CAN MAKE maxCoord TURN
             OUT TO BE DIFFERENT THAN THE "THEORETICAL" VALUE.
         */
      uLong             maxCoord ;
      
    } normInfoType ;

/* ########################################################## */
/*                       QheaderType                          */
/* ########################################################## */
/*
      A STRUCTURE FOR REPRESENTING A QUEUE, USING AN UNDERLYING
      LINKED LIST.  THE ACTUAL STORAGE SPACE FOR THE ELEMENTS
      OF THE QUEUE IS INTENDED TO BE IN AN ARRAY (THE pointList
      FIELD OF A VARIABLE OF ramDataRepType).  THE Qfront AND
      Qrear FIELDS ARE INTENDED TO BE INDEXES INTO THIS ARRAY,
      GIVING THE LOCATIONS OF THE FRONT AND REAR ELEMENTS OF
      THE QUEUE.  ANOTHER FIELD IN THE ramDataRepType VARIABLE,
      THE nextAfter ARRAY, IS USED TO MAKE THE LINKS FROM ONE
      ELEMENT OF THE QUEUE TO ITS SUCCESSOR.
*/      

typedef struct
{
   uLong  Qfront,  /* INDEXES USED AS POINTERS TO THE FRONT */
          Qrear;   /* AND REAR OF THE QUEUE */
   uLong  Qsize ;
} QheaderType ;

/* ########################################################## */
/*                     stackHeaderType                        */
/* ########################################################## */
/*
      A STRUCTURE FOR REPRESENTING A STACK, USING AN UNDERLYING
      LINKED LIST.  THE ACTUAL STORAGE SPACE FOR THE ELEMENTS
      OF THE STACK IS INTENDED TO BE IN AN ARRAY (THE pointList
      FIELD OF A VARIABLE OF ramDataRepType).  THE stkTop FIELD
      IS INTENDED TO BE AN INDEX INTO THIS ARRAY, GIVING THE
      LOCATIONS OF THE TOP ELEMENTS OF THE STACK.  ANOTHER
      FIELD IN THE ramDataRepType VARIABLE, THE nextAfter
      ARRAY, IS USED TO MAKE THE LINKS FROM ONE ELEMENT OF THE
      STACK TO ITS SUCCESSOR.
*/      

typedef struct
{
   uLong  stkTop ;  /* INDEX USED AS POINTER TO STACK TOP. */
   uLong  stkSize ; /* SIZE FIELD TO MAKE  */
} stackHeaderType;

/* ########################################################## */
/*                     ramDataRepType                         */
/* ########################################################## */
/*
      A STRUCTURE FOR THE INTERNAL REPRESENTATION OF AN OBJECT
      WHOSE FRACTAL DIMENSION IS TO BE ESTIMATED.  WE INCLUDE
      TOOLS FOR DOING THE VARIOUS MANIPULATIONS OF THE DATA
      THAT ARE NECESSARY.
*/      
  typedef struct
    {
          /*
             INFO ABOUT THE SIZE OF THE ATOMIC TYPES USED HERE.
          */
      int       ulongBitNum ;
          /*
             INFO ABOUT THE DIMENSIONS OF THE ARRAYS STORED
             HERE.          columns        rows
                            -------        ----
          */
      uLong    embedDim,     dataLines ;
         /*
             ARRAY OF TWO QUEUES FOR THE RADIX SORT
          */
      QheaderType          Q [2]   ;
          /*
             POINTER TO THE FREE LIST
          */
      stackHeaderType    avail  ;    
          /*
             THE DATA POINTS -- BASICALLY A RECTANGULAR ARRAY
          */ 
      uLong ** pointList ;
          /*
             AN ARRAY OF INTEGERS USED TO STORE POINTERS
             (LINKS) FROM ONE ELEMENT OF pointList TO ANOTHER.
             A POINTER IS IMPLEMENTED AS THE INDEX INTO
             pointList OF THE DESIRED ELEMENT (DATA POINT)
          */
      uLong *  nextAfter ;
    } ramDataRepType ;

/* ########################################################## */
/*                      coverDataType                         */
/* ########################################################## */
/*
      A STRUCTURE FOR STORING ATTRIBUTES OF COVERINGS OF A DATA
      SET.
*/      
  typedef struct
    {

/* ---------------------------------------------------------- */
/*                  Sweep Temp Variables                      */
/* ---------------------------------------------------------- */
          /* 
               THE NEXT FEW VARIABLES ARE KEPT HERE TO AVOID
               THE OVERHEAD OF ALLOCATING AND/OR INITIALIZING
               LOCAL VARIABLES IN SWEEP OPERATIONS THAT EXECUTE
               O(n) or O(nlog(n)) TIMES (n = SIZE OF THE INPUT
               FILE).  WE MAY ALSO WANT TO FIND A WAY TO CUT
               DOWN ON INDIRECT ADDRESSING DURING THE SWEEP,
               POSSIBLY BY MAKING A COPY OF THIS RECORD.
          */
          /*
               A FREQUENTLY USED CONSTANT.
          */
      double     log2 ;
          /*
             AN ARRAY OF INTEGERS, ONE FOR EACH COORDINATE
             POSITION OF THE DATA.  THE DIFFERENCE BETWEEN TWO
             DATA POINTS IS REPRESENTED BY THE ARRAY OF XOR'S
             OF THE RESPECTIVE COORDINATES.
          */
      uLong    *    ptDiff ;
          /*
               A FLAG FOR DETERMINING WHEN A BIT CHANGE HAS BEEN
               DETECTED BETWEEN TWO POINTS.
          */
      int        bitChFnd ;
          /*
               AN INTEGER FIELD TO RECORD THE POSITION OF THE
               LAST BIT CHANGE.
          */
      int        MSBdiff ;
          /*
               TEMP INDEX VARIABLES FOR COORDINATE AND BIT
               POSITIONS.
          */
      uLong      coord ;
      uLong      bitPos ;
          /*
               TEMP STORAGE FOR A SAMPLE PROBABILITY THAT IS
               USED TO CALCULATE OTHER STATISTICS.
          */
      double     sProb ; 
          /*
               THE NUMBER OF CELL SIZES USED IN THE COVERING
               CALCULATIONS.  THIS IS ALSO THE SIZE OF THE
               ARRAYS STORED HERE.  
          */
      uLong        numCellSizes ;
            /*
               THE RATIO OF SUCCESSIVE CELL SIZES. 
            */
      double     magnifier ;
         /*
             THIS IS THE LARGEST VALUE THAT WILL BE FOUND AMONG
             THE COORDINATES OF THE NORMALIZED SET OF DATA
             POINTS.  IN "THEORY", THE VALUE OF maxCoord SHOULD
             BE THE SAME AS THE LARGEST POSSIBLE VALUE OF AN
             UNSIGNED LONG INT.  HOWEVER, THERE IS SOME
             INSTABILITY INHERENT IN THE WAY THAT THE SCALING
             FACTOR IS CALCULATED.  THIS CAN MAKE maxCoord TURN
             OUT TO BE DIFFERENT THAN THE "THEORETICAL" VALUE.
         */
      uLong             maxCoord ;
          /*
             THE SIZE OF THE EMBEDDING SPACE -- USED AS AN
             UPPER BOUND TO THE FRACTAL DIMENSION BY A FUNCTION
             THAT ESTIMATES THE SATURATION POINT OF THE DATA
             SET.
          */
      uLong      embedDim ;
          /*
             AN ARRAY FOR STORING THE CARDINALITY OF
             COVERINGS COMPOSED OF CELLS OF A GIVEN FIXED SIDE
             LENGTH.  THE INDEX OF THE ARRAY IS IS THE LOG
             (BASE 2) OF THE LENGTH OF THE SIDE OF THE CELL IN
             THE COVERING.
          */
      uLong    *    cellCount ;
          /*
             totDistPts IS THE TOTAL NUMBER OF DISTINGUISHABLE
             POINTS IN THE INPUT SET.  WE INTEND TO USE
             totDistPts ONLY WITH THE MINCOVER (-m) OPTION.
             WHEN WE ESTIMATE WITH THE LIEBOVITCH-TOTH METHOD,
             WE USE THE COUNT OF CELLS OF SIZE ONE INSTEAD.
          */
      uLong         totDistPts ;
          /*
             A COUNTER FOR EACH CELL SIZE.  THE COUNTERS ARE
             USED BY SweepData TO KEEP TRACK OF THE NUMBER OF
             DATA POINTS IN THE "CURRENT" CELL OF EACH SIZE.
             SweepData CHECKS THE COUNTS INTERMITTANTLY IN
             ORDER TO DERIVE OTHER STATISTICS -- THOSE NEEDED
             FOR THE ESTIMATION OF INFORMATION AND CORRELATION
             DIMENSION.
          */
      uLong    *    pointCount ;
          /*
             AN ARRAY FOR STORING THE BASE 2 LOGARITHMS OF THE
             SIDE LENGTHS OF CELLS WHICH COMPOSE A COVERING OF
             THE DATA.    
          */
      double    *    logCellSize ;
          /*
             VARIOUS STATISTICS DERIVED FROM PROPERTIES OF THE
             COVERINGS.
          */
      double              *    negLogCellCount ;
      double              *    information     ;
      double              *    sumSqrFreq      ;
      double              *    logSumSqrFreq   ;
      uLong                    saturationPoint ;
      uLong                    fineMeshPoint   ;
      
    } coverDataType ;
/* ########################################################## */
/*                    dimEstimDataType                        */
/* ########################################################## */
/*
      A STRUCTURE FOR STORING THE RESULTS OF THE PROCESS OF
      FRACTAL DIMENSION ESTIMATION.  IT IS EXPECTED THAT SOME
      PROCEDURE WILL INPUT coverData, AND OUT PUT DATA OF THIS
      TYPE.  FIELDS OF THIS STRUCTURE WILL INCLUDE ACTUAL
      DIMENSION ESTIMATES, AS WELL AS RELATED INFORMATION NOT
      CONSIDERED TO BE RAW "coverData".
*/      
  typedef struct
    {
          /*
             INDEXES MARKING A REGION OF THE coverData
             CONSIDERED TO BE ACCEPTABLE AS INPUT INTO THE
             ESTIMATION PROCEDURE.
          */
      uLong    saturationPoint ;
      uLong    fineMeshPoint   ;
          /*
             DIMENSION ESTIMATES COMPUTED AS SLOPES OF LEAST
             SQUARE LINES FITTED TO PARTS OF coverData.  
          */
      double               capDimEst       ;
      double               infDimEst       ;
      double               corrDimEst      ;
          /*
             INTERCEPTS CORRESPONDING TO THE SLOPES ABOVE.
             HAVING THIS DATA HERE MEANS WE HAVE ENOUGH
             INFORMATION TO DETERMINE THE LEAST SQUARE LINES
             USED IN THE ESTIMATION OF THE DIMENSIONS.
          */
      double               capIntercept    ;
      double               infIntercept    ;
      double               corrIntercept   ;
      
          /*
             APPROXIMATE R-SQUARE ERROR TERMS TO MEASURE THE
             GOODNESS OF FIT OF THE LEAST SQUARES LINE.  THIS
             MUST NOT BE TAKEN AS A MEASURE OF THE ACCURACY OF
             THE DIMENSION ESTIMATE.  IT SIMPLY INDICATES HOW
             WELL THE LINE FITS THE LOG-LOG PLOT.
          */

      double               capRSqr       ;
      double               infRSqr       ;
      double               corrRSqr      ;

    } dimEstimDataType ;

/* ########################################################## */
/*                     refinedCDtype                          */
/* ########################################################## */
/*
    THIS IS A DATA STRUCTURE FOR STORING COVERING STATISTICS
    GLEANED FROM SEVERAL SWEEPS OF THE SAME DATA, USING
    DIFFERENT SCALINGS FOR EACH SWEEP.
*/
typedef struct
{
    uLong      embedDim        ;
    int        refDepth        ;
    int        ulongBitNum     ;
    uLong      denom           ;
    uLong      numSets         ;
    uLong      numCellSizes    ;   
    uLong   *  scaleIndex      ;   /*  ARRAY  */
    double  *  logCellSize     ;   /*  ARRAY  */
    uLong   *  cellCount       ;   /*  ARRAY  */
    double  *  negLogCellCount ;   /*  ARRAY  */
    double  *  information     ;   /*  ARRAY  */
    double  *  logSumSqrFreq   ;   /*  ARRAY  */
    uLong      saturationPoint ;
    uLong      fineMeshPoint   ;
} refinedCDtype ;

/* ######################################################### */
/* ######################################################### */
/* ######################################################### */     
/*                        end fd.h                           */
/* ######################################################### */
/* ######################################################### */
/* ######################################################### */


