
/* BEGIN COPYRIGHT NOTICE

Copyright (c) 1992 by John Sarraille and Peter DiFalco
(john@ishi.csustan.edu)

Copyright (c) 1998 by John Sarraille
(john@ishi.csustan.edu)

Permission to use, copy, modify, and distribute this software
and its documentation for any purpose and without fee is hereby
granted, provided that the above copyright notice appear in all
copies and that both that copyright notice and this permission
notice appear in supporting documentation.

The algorithm used in this program was inspired by the paper
entitled "A Fast Algorithm To Determine Fractal Dimensions By
Box Counting", which was written by Liebovitch and Toth, and
which appeared in the journal "Physics Letters A", volume 141,
pp 386-390, (1989).

This program is not warranted: use at your own risk.

END COPYRIGHT NOTICE */


/*

   FD3 PROGRAM FILE

Filename:   optm.c

OUTLINE OF CALLS:

mOptDimErr +
mOptArgs +
  IsOptionString
  GetNumSizes +
    GetIntFromString
  MagnifLimit +
  GetMagnifier +
  CellSizesLimit +

*/

#include "fd.h"

/* 
    THE FUNCTIONS IN THIS FILE: 
*/
void   mOptDimErr     () ;
int    mOptArgs       () ;
double MagnifLimit    () ;
int    CellSizesLimit () ;
int    GetMagnifier   () ;
int    GetNumSizes    () ;

/*   
      THE FUNCTIONS NOT IN THIS FILE,
      BUT CALLED FROM HERE:
*/
extern int      IsOptionString       () ;     /*  in convert.c    */
extern int      GetULfromString      () ;     /*  in convert.c    */

/* ######################################################### */
/*                         mOptDimErr                        */
/* ######################################################### */
/*
     THIS IS AN ERROR MESSAGE THAT IS PRINTED IF THE USER
     ATTEMPTS TO USE THE MINCOVER OPTION WHEN THE EMBEDDING
     DIMENSION IS GREATER THAN ONE.
*/

void mOptDimErr(fRepPtr) 
  fileRepType * fRepPtr ;
{
  printf("File \"%s\" appears to contain a set\n", fRepPtr->fileName) ;
  printf("with embedding dimension equal to %lu.\n", fRepPtr->embedDim) ;
  printf("With the -m (mincover) option set, fd3 can process\n") ;
  printf("only files with embedding dimension equal to 1.\n") ;
  printf("\n") ;
  printf("Sorry, fd3 must abort now.\n") ;
  printf("\n") ;
  fflush(stdout); exit(EXIT_FAILURE) ;
}
/* ######################################################### */
/*                         mOptArgs                          */
/* ######################################################### */
/*
     THIS FUNCTION TAKES CARE OF INPUTTING THE COMMAND-LINE
     ARGUMENTS, IF ANY, THAT ACCOMPANY THE -m OPTION.  THE
     ARGUMENTS ARE A (double) MAGNIFIER AND A (integer) NUMBER
     OF CELL SIZES.  IF EITHER ARGUMENT IS GIVEN, THEN BOTH
     MUST BE GIVEN.  HOWEVER, THE LETTER 'd' MAY BE USED
     INSTEAD OF A NUMBER FOR EITHER OF THE OPTIONS, IN ORDER TO
     GET A DEFAULT VALUE.

*/
int mOptArgs(argsLftPt, curArgPt, argv, cPt, mPt) 
       int * argsLftPt ;
       int * curArgPt ;
       char * argv[] ;
       commandType * cPt ;
       machineType * mPt ;
{
   int syntaxOK = TRUE;
   int magIsDefault, cSizesIsDefault ;
   
        /* IF THERE ARE TWO OR MORE STRINGS TO THE RIGHT OF "-m"
           (BESIDES THE FILENAME) AND IF THE NEXT STRING AFTER
           "-m" IS NOT AN OPTION STRING ... */
  if ( ((* argsLftPt) >= 2)
       && (!IsOptionString(argv[(* curArgPt)])))
           /*
              THEN, ASSUMING CORRECT SYNTAX, THE NEXT TWO
              STRINGS AFTER "-m" MUST BE ARGUMENTS FOR "-m", SO
              PARSE THEM, AND ACCORDINGLY UPDATE THE NUMBER OF
              ARGUMENTS LEFT TO PROCESS AND THE INDEX OF THE
              NEXT ARGUMENT TO PROCESS.
           */ 
       {
           /*
               WE NEED A LITTLE SPECIAL HANDLING IF ONE OR BOTH
               OF THE ARGUMENTS TO THE -m OPTION IS A 'd'.  THIS
               DENOTES THAT THE USER WANTS A DEFAULT VALUE.
           */
         magIsDefault = (strlen(argv[(* curArgPt)])==1)
                             && (argv[(* curArgPt)][0]=='d') ;
         cSizesIsDefault = (strlen(argv[(* curArgPt)+1])==1)
                             && (argv[(* curArgPt)+1][0]=='d') ;

          if (magIsDefault)
          { if (!cSizesIsDefault)
                 /*
                    GET THE NUMBER OF CELL SIZES FROM THE
                    SECOND STRING AND CREATE A CORRESPONDING
                    MAGNIFIER.
                 */
            { syntaxOK = (GetNumSizes(argv[(* curArgPt)+1], cPt, mPt));
              if (syntaxOK)
                  (cPt->magnifier) =
                     MagnifLimit(cPt->numCellSizes, mPt->maxUlong) ;
            }
          }
          else
          {
            if (cSizesIsDefault)
                 /*
                    GET THE MAGNIFIER FROM THE FIRST STRING AND
                    CREATE A CORRESPONDING NUMBER OF CELL
                    SIZES.
                 */
            { syntaxOK = (GetMagnifier(argv[(* curArgPt)], cPt, mPt));
              if (syntaxOK)
                  (cPt->numCellSizes) =
                     CellSizesLimit
                      (cPt->magnifier, mPt->maxUlong, mPt->maxInt) ;
            }
            else
                       /*
                          NEITHER STRING IS A 'd' SO GET BOTH
                          ARGUMENTS FROM THE STRINGS.
                       */
              syntaxOK = (GetMagnifier(argv[(* curArgPt)], cPt, mPt))
                     && (GetNumSizes(argv[(* curArgPt)+1], cPt, mPt)) ;
          }
         (* curArgPt) =  (* curArgPt) + 2 ;
         (* argsLftPt) = (* argsLftPt) - 2 ;
       }
  return(syntaxOK) ;
}
/* ######################################################### */
/*                        MagnifLimit                        */
/* ######################################################### */
/*
    CALCULATE THE MAXIMUM MAGNIFIER THAT CAN BE USED WITH THE
    GIVEN NUMBER OF CELL SIZES AND THE GIVEN BOUNDING CELL
    DIAMETER.  IT WORKS LIKE THIS: THE BOUNDING CELL IS THE
    LARGEST CELL SIZE, AND EACH SUCCESSIVE CELL SIZE IS
    OBTAINED BY DIVIDING BY THE MAGNIFIER.  WE STOP AT THE LAST
    CELL SIZE GREATER THAN ONE, SINCE THE SCALING WE ARE USING
    ASSURES THAT A CELL OF SIZE ONE OR LESS CAN ONLY CONTAIN
    ONE POINT.
*/
double MagnifLimit (numCellSizes, bCellDiam)
       uLong numCellSizes ;
       uLong bCellDiam ;
{
  double bCD = (double) bCellDiam ;
  uLong  n = numCellSizes - 1 ;

  if (n < 1 ) n = 1 ;  /* just in case... */
     /* RETURN THE n'th ROOT OF THE DIAMETER.*/
  return(exp(log(bCD)/(n))) ; 
}
/* ######################################################### */
/*                      CellSizesLimit                       */
/* ######################################################### */
/*
    CALCULATE THE MAXIMUM NUMBER OF CELL SIZES THAT CAN BE USED
    WITH THE GIVEN MAGNIFIER AND THE GIVEN BOUNDING CELL
    DIAMETER.  IT WORKS LIKE THIS: THE BOUNDING CELL IS THE
    LARGEST CELL SIZE, AND EACH SUCCESSIVE CELL SIZE IS
    OBTAINED BY DIVIDING BY THE MAGNIFIER.  WE STOP AT THE LAST
    CELL SIZE GREATER THAN ONE, SINCE THE SCALING WE ARE USING
    ASSURES THAT A CELL OF SIZE ONE OR LESS CAN ONLY CONTAIN
    ONE POINT.

    BASICALLY, THE ANSWER IS "THE BIGGEST INTEGER NO GREATER
    THAN LOG (BASE=MAGNIFIER) OF THE BOUNDING CELL DIAMETER".
    HOWEVER, IF THE MAGNIFIER IS 1 + EPSILON, THEN THE NUMBER
    OF POSSIBLE CELL SIZES GOES TO INFINITY AS EPSILON GOES TO
    ZERO.  WE THEREFORE LIMIT THE NUMBER OF CELL SIZES TO max.
*/
int CellSizesLimit (magnifier, bCellDiam, max)
       double  magnifier ;
       uLong bCellDiam ;
       int max ;
{
   double bCD = (double) bCellDiam ;
   
   if   ( (magnifier <= 1) ||
            floor(log(bCD)/log(magnifier))+1 > max )
        return(max) ;
   else return((int) floor(log(bCD)/log(magnifier))+1) ;
}
/* ######################################################### */
/*                        GetMagnifier                       */
/* ######################################################### */
/*
    GET THE NUMBER THAT DETERMINES THE RATIO OF SUCCESSIVE
    CELL SIZES WHEN THE MIN-COVER VERSION OF THE PROGRAM RUNS.

    THE BIGGEST VALUE ONE CAN CHOOSE FOR THE MAGNIFIER, WITHOUT
    BEING COMPLETELY SILLY, IS THE SIZE OF THE BOUNDING CELL
    FOR THE SCALED DATA: maxUlong.  THIS MAGNIFIER LIMITS US TO
    HAVING ONLY TWO CELL SIZES: ONE SIZE THAT COVERS ALL THE
    DATA, AND ONE SIZE THAT CAN ONLY HOLD A SINGLE POINT.  THIS
    IS STILL PRETTY RIDICULOUS, BUT WE USE THIS LIMIT TO ALLOW
    THE WIDEST LATITUDE FOR EXPERIMENTATION.
*/
int GetMagnifier (strPt, cPt, machPt)
       char *         strPt ;
       commandType *  cPt ;
       machineType * machPt ;
{
  int result ;
  double tempValue;
  uLong max = (machPt->maxUlong) ;

  result = sscanf(strPt,"%lf", &tempValue);
      /*
         FIRST TRY TO DETERMINE IF THE STRING WAS SUCCESSFULLY
         CONVERTED.  IF SO, THEN THE VALUE IS PROBABLY OK AS
         LONG AS IT IS WITHIN A CERTAIN RANGE.
      */
  if ((result != 0) && (tempValue > 1.0) && (tempValue <= max))
     {
       (cPt->magnifier) =  tempValue;
       return (TRUE) ;
     }
  else return (FALSE) ;
}
/* ######################################################### */
/*                        GetNumSizes                        */
/* ######################################################### */
/*
   GET THE NUMBER OF CELL SIZES FOR THE MIN-COVER VERSION OF
   THE PROGRAM.
*/
int GetNumSizes (strIntPtr, cPt, mPt)
       char *         strIntPtr ;
       commandType *  cPt ;
       machineType *  mPt ;
{  
   int   valueOK ;
   uLong value, ulongMax ;

   ulongMax = (mPt->maxUlong) ;
   valueOK = GetULfromString (strIntPtr, ulongMax, &value) ;
   if ((valueOK) && (value > 1))
      {
         (cPt->numCellSizes) = value ;
         return (TRUE) ;
      }
   else return (FALSE) ;
}
/* ######################################################### */
/* ######################################################### */
/* ######################################################### */     
/*                        end optm.c                         */
/* ######################################################### */
/* ######################################################### */
/* ######################################################### */
