
/* BEGIN COPYRIGHT NOTICE

Copyright (c) 1992 by John Sarraille and Peter DiFalco
(john@ishi.csustan.edu)

Copyright (c) 1998 by John Sarraille
(john@ishi.csustan.edu)

Permission to use, copy, modify, and distribute this software
and its documentation for any purpose and without fee is hereby
granted, provided that the above copyright notice appear in all
copies and that both that copyright notice and this permission
notice appear in supporting documentation.

The algorithm used in this program was inspired by the paper
entitled "A Fast Algorithm To Determine Fractal Dimensions By
Box Counting", which was written by Liebovitch and Toth, and
which appeared in the journal "Physics Letters A", volume 141,
pp 386-390, (1989).

This program is not warranted: use at your own risk.

END COPYRIGHT NOTICE */


/*

   FD3 PROGRAM FILE

Filename:  queue.c

OUTLINE OF CALLS:

CreateStk +
ClearStk +
EmptyStk +
PushStk +
PopStk +
InitPushStk +
PrintStk +
CreateQ +
ClearQ +
EmptyQ +
SizeQ +
FrontIndex +
RearIndex +
Enq +
EnqFast +
Deq +
DeqFast +
TransfQ +
PrintQ +
PrintQmesh +
ConcatQs +

*/

#include "fd.h"

      /* THE FUNCTIONS IN THIS FILE: */

void CreateStk      () ;
void ClearStk       () ;
int  EmptyStk       () ;
void PushStk        () ;
void PopStk         () ;
void InitPushStk    () ;
void PrintStk       () ;

void  CreateQ        () ;
void  ClearQ         () ;
int   EmptyQ         () ;
uLong SizeQ          () ;
uLong FrontIndex     () ;
uLong RearIndex      () ;
void  Enq            () ;
void  EnqFast        () ;
void  Deq            () ;
void  DeqFast        () ;
void  TransfQ        () ;
void  PrintQ         () ;
void  PrintQmesh     () ;
void  ConcatQs       () ;

/* ######################################################### */
/*
   THE PURPOSE OF THIS FILE IS MAINLY TO IMPLEMENT A SET OF
   FUNCTIONS TO DO OPERATIONS ON A QUEUE.  HOWEVER, THERE IS
   SOME POSSIBLE NEED TO MAINTAIN A LIST OF FREE RECORDS, I.E.
   RECORDS WHICH ARE NOT IN ANY QUEUE.  A STACK IS A CONVENIENT
   DATA STRUCTURE FOR THIS PURPOSE.  THE NEXT FEW FUNCTION
   DEFINITIONS ARE THE IMPLEMENTATION OF THE STACK OPERATIONS.
*/   
/* ######################################################### */
/*                         CreateStk                         */
/* ######################################################### */
/*
    CREATES AN EMPTY STACK.
*/
void CreateStk(stackPtr)
      stackHeaderType  * stackPtr;
{
   (stackPtr->stkSize) = 0 ;
   return ;
}
/* ######################################################### */
/*                          ClearStk                         */
/* ######################################################### */
/*
    CLEARS (EMPTIES) THE STACK.
*/
void ClearStk(stackPtr)
      stackHeaderType  * stackPtr;
{
   (stackPtr->stkSize) = 0 ;
   return ;
}
/* ######################################################### */
/*                          EmptyStk                         */
/* ######################################################### */
/*
    A SIMPLE TEST TO TELL IF A STACK IS EMPTY.
*/

int EmptyStk (stackPtr)
      stackHeaderType  * stackPtr;
{
  return ( (stackPtr->stkSize) == 0);
}
/* ######################################################### */
/*                          PushStk                          */
/* ######################################################### */
/*
    "eltIndex" IS AN INDEX INTO AN ARRAY.  IT CAN BE ANY ARRAY.
    THIS PROCEDURE, IN EFFECT, PUTS ONTO THE TOP OF A STACK THE
    ARRAY DATA TO WHICH THE INDEX REFERS.  THERE IS NO ACTUAL
    MOVEMENT OF THE ARRAY DATA.  THE EFFECT IS ACHIEVED BY
    ADJUSTING A VALUE IN AN INDEX ARRAY, BY CHANGING THE INDEX
    REPRESENTING THE LOCATION OF THE TOP OF THE STACK, AND BY
    INCREMENTING THE stkSize VARIABLE.  CAUTION: NO OVERFLOW
    CHECKING IS DONE.  THE CALLER IS RESPONSIBLE FOR ASSURING
    THAT THERE IS REALLY "ROOM" FOR THE ELEMENT -- I.E. THAT
    ARRAY BOUNDS AND VARIABLE STORAGE LIMITS ARE NOT EXCEEDED.
*/
void PushStk(stackPtr, indexArray, eltIndex)
       stackHeaderType  * stackPtr;
       uLong * indexArray, eltIndex ;
{
  indexArray[eltIndex] = (stackPtr->stkTop) ;
  (stackPtr->stkTop) = eltIndex ;
  (stackPtr->stkSize)++ ;
  return ;
}
/* ######################################################### */
/*                           PopStk                          */
/* ######################################################### */
/*
    THIS PROCEDURE IS COMPLEMENTARY TO PushStk ABOVE.  IT
    REMOVES AN ELEMENT FROM THE STACK AND RETURNS A POINTER TO
    THE INDEX OF THE POPPED ELEMENT.  CAUTION: NO UNDERFLOW
    CHECKING IS DONE.  THE CALLER IS RESPONSIBLE FOR ASSURING
    THAT THE STACK IS NOT EMPTY.
*/
void PopStk(stackPtr, indexArray, eltIndexPtr)
      stackHeaderType  * stackPtr;
      uLong * indexArray, * eltIndexPtr ;
{
   *eltIndexPtr = (stackPtr->stkTop) ;
   (stackPtr->stkTop) = indexArray[(stackPtr->stkTop)] ;
   (stackPtr->stkSize)-- ;
   return ;
}
/* ######################################################### */
/*                       InitPushStk                         */
/* ######################################################### */
/*
    A PROCEDURE TO EFFICIENTLY INITIALIZE A STACK TO CONTAIN
    THE ELEMENTS OF AN ARRAY.
*/
void InitPushStk(stackPtr, indexArray, length)
      stackHeaderType  * stackPtr;
      uLong * indexArray, length ;

{
  uLong i;
  for (i=0; i < length-1; i++) indexArray[i] = i+1;
  (stackPtr->stkTop) = 0 ;
  (stackPtr->stkSize) = length ;
  return ;
}
/* ######################################################### */
/*                         PrintStk                          */
/* ######################################################### */
/*
    A PROCEDURE TO PRINT A STACK.  IT IS INTENDED TO BE USED AS
    AN AID TO DEBUGGING.  THIS IS A SPECIALIZED PROCEDURE THAT
    NEEDS TO USE INFORMATION ABOUT THE IMPLEMENTATION OF THE
    RECORDS STORED IN THE STACK.  
*/
void PrintStk(stackPtr, indexArray, dataArray, width)
      stackHeaderType  * stackPtr;
      uLong * indexArray, ** dataArray, width;
{
  uLong tempIndex, count ;
  long int coord;
  printf("Stack Top is: %lu.  ", stackPtr->stkTop);
  printf("Stack Size is: %lu.\n\n", stackPtr->stkSize);
  printf("Index\tNext Index\tCoords\n\n");
  tempIndex = (stackPtr->stkTop); count = 0;
  while (count < (stackPtr->stkSize))
    {
        printf("%lu\t%lu\t", tempIndex, indexArray[tempIndex]);
        for (coord = 0 ; coord < width; coord++)
          printf("%12lu", dataArray[coord][tempIndex]);
        printf("\n");
        count++ ;  tempIndex = indexArray[tempIndex];
     }
  printf("\n");
  return ;     
}

/* ######################################################### */
/*
   THE NEXT FEW FUNCTION DEFINITIONS ARE THE IMPLEMENTATION OF
   THE QUEUE OPERATIONS.
*/
/* ######################################################### */


/* ######################################################### */
/*                            CreateQ                        */
/* ######################################################### */
/*
     A PROCEDURE TO INITIALIZE A QUEUE.  IT DOES *NOT* ALLOCATE
     STORAGE FOR THE QUEUE.
*/
void CreateQ (queuePtr)
       QheaderType * queuePtr ;
{
   (queuePtr->Qsize) = 0 ;
   return ;
}
/* ######################################################### */
/*                          ClearQ                           */
/* ######################################################### */
/*
     A PROCEDURE TO CLEAR (EMPTY) A QUEUE.
*/
void ClearQ (queuePtr)
       QheaderType * queuePtr ;
{
   (queuePtr->Qsize) = 0 ;
   return ;
}
/* ######################################################### */
/*                         EmptyQ                            */
/* ######################################################### */
/*
    A TEST TO TELL IF A QUEUE IS EMPTY.
*/
int EmptyQ(queuePtr)
     QheaderType * queuePtr ;
{
   return( (queuePtr->Qsize) == 0);
}
/* ######################################################### */
/*                         SizeQ                             */
/* ######################################################### */
/*
    A FUNCTION TO RETURN THE SIZE OF A QUEUE.
*/
uLong SizeQ (queuePtr)
     QheaderType * queuePtr ;
{
   return(queuePtr->Qsize);
}
/* ######################################################### */
/*                        FrontIndex                         */
/* ######################################################### */
/*
    A FUNCTION TO RETURN THE INDEX OF THE FRONT ELEMENT OF A
    QUEUE.
*/
uLong FrontIndex (queuePtr)
     QheaderType * queuePtr ;
{
   return(queuePtr->Qfront);
}
/* ######################################################### */
/*                         RearIndex                         */
/* ######################################################### */
/*
    A FUNCTION TO RETURN THE INDEX OF THE REAR ELEMENT OF A
    QUEUE.
*/
uLong RearIndex (queuePtr)
     QheaderType * queuePtr ;
{
   return(queuePtr->Qrear);
}
/* ######################################################### */
/*                            Enq                            */
/* ######################################################### */
/*
   "eltIndex" IS AN INDEX INTO AN ARRAY.  IT CAN BE ANY ARRAY.
    THIS PROCEDURE, IN EFFECT, PUTS ONTO THE REAR OF A QUEUE THE
    ARRAY DATA TO WHICH THE INDEX REFERS.  THERE IS NO ACTUAL
    MOVEMENT OF THE ARRAY DATA.  THE EFFECT IS ACHIEVED BY
    ADJUSTING LINKS IMPLEMENTED AS ARRAY INDEXES.  CAUTION:
    THIS PROCEDURE DOES NO "OVERFLOW" CHECKING.  THE CALLER
    MUST ENSURE THAT THE QUEUE IS NOT FULL.  THIS PROCEDURE
    ALSO DOES NOT REMOVE THE ENQUEUED OBJECT FROM ANY "FREE
    LIST" THAT MIGHT CONTAIN IT.
*/
void Enq (queuePtr, indexArray, eltIndex)
       QheaderType * queuePtr ;
       uLong * indexArray, eltIndex ;
{
  if    ( (queuePtr->Qsize) == 0 )
        (queuePtr->Qfront) = eltIndex ;
  else  indexArray [queuePtr->Qrear] = eltIndex ;
  (queuePtr->Qrear) = eltIndex ;
  (queuePtr->Qsize)++ ;
  return ;
}
/* ######################################################### */
/*                         EnqFast                           */
/* ######################################################### */
/*
    THIS FAST ENQUEUE OPERATION IS LIKE "Enq", EXCEPT THAT IT
    ASSUMES THAT THE QUEUE IS NOT EMPTY AT THE START, AND SAVES
    TIME BY NOT TESTING THIS CONDITION.
*/
void EnqFast (queuePtr, indexArray, eltIndex)
       QheaderType * queuePtr ;
       uLong * indexArray, eltIndex ;
{
  indexArray [queuePtr->Qrear] = eltIndex ;
  (queuePtr->Qrear) = eltIndex ;
  (queuePtr->Qsize)++ ;
  return ;
}
/* ######################################################### */
/*                              Deq                          */
/* ######################################################### */
/*
   "eltIndexPtr" IS A POINTER TO A DATA STRUCTURE SUITABLE FOR
    STORING AN ARRAY INDEX.  IT CAN BE ANY ARRAY.  THIS
    PROCEDURE, IN EFFECT, REMOVES ONE ELEMENT OF THE ARRAY FROM
    THE FRONT OF A QUEUE AND WRITES THE INDEX OF THE ELEMENT IN
    THE STORAGE CELL REFERRED TO BY eltIndexPtr.  THERE IS NO
    ACTUAL MOVEMENT OF ANY ARRAY DATA.  THE EFFECT IS ACHIEVED
    BY ADJUSTING LINKS IMPLEMENTED AS ARRAY INDEXES.  CAUTION:
    THIS PROCEDURE DOES NO "UNDERFLOW" CHECKING.  THE CALLER
    MUST ENSURE THAT THE QUEUE IS NOT EMPTY.  THIS PROCEDURE
    ALSO DOES NOT PLACE THE DEQUEUED OBJECT ON A "FREE LIST".
*/
void Deq (queuePtr, indexArray, eltIndexPtr)
       QheaderType * queuePtr ;
       uLong * indexArray, * eltIndexPtr ;
{
  *eltIndexPtr = (queuePtr->Qfront) ;
  (queuePtr->Qfront) = indexArray[queuePtr->Qfront];
  (queuePtr->Qsize)-- ;
  return ;
}
/* ######################################################### */
/*                           DeqFast                         */
/* ######################################################### */
/*
    THIS PROCEDURE IS MEANT TO BE A FAST DEQUEUE THAT WE CAN
    USE IF THE QUEUE CONTAINS MORE THAN ONE ELEMENT.  SINCE, IN
    THE PRESENT QUEUE IMPLEMENTATION, WE ARE KEEPING A COUNTER
    OF THE NUMBER OF ELEMENTS IN EACH QUEUE, WE DON'T NEED TO
    FLAG AN EMPTY QUEUE WITH NIL FRONT OR REAR POINTERS, SO THE
    CODE HERE IS THE SAME AS FOR THE (REGULAR) Enq.

*/
void DeqFast (queuePtr, indexArray, eltIndexPtr)
       QheaderType * queuePtr ;
       uLong * indexArray, * eltIndexPtr ;
{
  *eltIndexPtr = (queuePtr->Qfront) ;
  (queuePtr->Qfront) = indexArray[queuePtr->Qfront];
  (queuePtr->Qsize)-- ;
  return ;
}
/* ######################################################### */
/*                          TransfQ                          */
/* ######################################################### */
/*
    THIS IS A FUNCTION WHICH MOVES AN "ELEMENT" DIRECTLY FROM A
    SOURCE QUEUE TO A TARGET QUEUE.  THE EFFECT IS USUALLY THE
    SAME AS IT WOULD BE IF CALLS TO Enq AND Deq WERE USED TO DO
    THIS TRANSFER, BUT THE OVERHEAD OF PROCEDURE CALLS IS
    HALVED.  THIS PROCEDURE IS IMPORTANT MAINLY BECAUSE IT WILL
    BE GENERALIZED TO DEVELOP A PROCEDURE WHICH PERFORMS AN
    ENTIRE PASS OF A RADIX SORT WITH ONLY ONE CALL.

    USE WITH EXTREME CAUTION.  THE EFFECT OF THE FUNCTION WILL
    BE INCORRECT IF THE TARGET QUEUE IS EMPTY!
*/
void TransfQ(srcQueuePtr, targQueuePtr, indexArray)
       QheaderType * srcQueuePtr, * targQueuePtr ;
       uLong * indexArray ;
{
  uLong tempIndex ;
      /*
           FAST DEQUEUE
      */
  tempIndex = (srcQueuePtr->Qfront) ;
  (srcQueuePtr->Qfront) = indexArray[srcQueuePtr->Qfront];
  (srcQueuePtr->Qsize)-- ;  
      /*
           FAST ENQUEUE
      */
  indexArray [targQueuePtr->Qrear] = tempIndex ;
  (targQueuePtr->Qrear) = tempIndex ;
  (targQueuePtr->Qsize)++ ;

  return ;
}
/* ######################################################### */
/*                           PrintQ                          */
/* ######################################################### */
/*
    A PROCEDURE TO PRINT A QUEUE.  IT IS INTENDED TO BE USED AS
    AN AID TO DEBUGGING.  THIS IS A SPECIALIZED PROCEDURE THAT
    NEEDS TO USE INFORMATION ABOUT THE IMPLEMENTATION OF THE
    RECORDS STORED IN THE QUEUE.
*/
void PrintQ(qPtr, indexArray, dataArray, width)
     QheaderType * qPtr ;
     uLong * indexArray, ** dataArray, width;
{
   uLong coord, tempIndex, nodeNum=0 ;
   printf("Queue front is: %lu.  ", qPtr->Qfront);
   printf("Queue rear is: %lu.  ", qPtr->Qrear);
   if    ((qPtr->Qsize) == 0)
     printf("Queue size is: %lu -- THE QUEUE IS EMPTY.\n\n", qPtr->Qsize);
   else
   {
     printf("Queue size is: %lu.\n\n", qPtr->Qsize);
     printf(" Index  Next    Coords\n\n");
     tempIndex = (qPtr->Qfront) ;
     while (nodeNum < (qPtr->Qsize)-1)
     {
       printf("%5lu%5lu", tempIndex, indexArray[tempIndex]);
       for (coord = 0; coord < width; coord++)
           printf("%12lu", dataArray[coord][tempIndex]);
       printf("\n");
       nodeNum++ ;  tempIndex = indexArray[tempIndex];
     }
       /*
           WE ARE USING UNSIGNED LONG INTS FOR THE indexArray
           ELEMENTS, SO THERE IS NO SPECIAL VALUE LIKE -1
           AVAILABLE TO SIGNIFY END-OF-QUEUE.  THEREFORE, WE
           ARRANGE BELOW TO PRINT "n/a" FOR THE VALUE OF THE
           LAST INDEX.
       */
     printf("%5lu%5s", tempIndex, "n/a");
     for (coord = 0; coord < width; coord++)
           printf("%12lu", dataArray[coord][tempIndex]);
     printf("\n");
     printf("\n");
   }
   return ;
}
/* ######################################################### */
/*                         PrintQmesh                        */
/* ######################################################### */
/*
    A PROCEDURE TO PRINT THE "MESH" OF THE POINTS IN A QUEUE.
    THIS MESH IS THE SORT KEY USED IN THE RADIX SORT.  FOR
    EXAMPLE, IF WE HAVE TWO BINARY NUMBERS X = (111) AND Y =
    (000), THEIR MESH IS 101010.  LIKE PrintQ, THIS ROUTINE IS
    INTENDED TO BE USED AS AN AID TO CHECKING THE CORRECTNESS
    OF THE PROGRAM.  

    IT IS A SPECIALIZED PROCEDURE THAT NEEDS TO USE INFORMATION
    ABOUT THE IMPLEMENTATION OF THE RECORDS STORED IN THE
    QUEUE.  IT ALSO EMPLOYS BIT-ORIENTED MACRO'S TO COMPUTE THE
    MESH.
    
*/
void PrintQmesh(queuePtr, indexArray, dataArray, numCoordsInPt, numBitsInCoord)
     QheaderType * queuePtr ;
     uLong * indexArray, ** dataArray, numCoordsInPt;
     int numBitsInCoord ;
{
   uLong coord, tmpIdx, nodeNum=0 ;
   int bitNum ;

   printf("Queue front is: %lu. ", queuePtr->Qfront);
   printf("Queue rear is: %lu.  ", queuePtr->Qrear);
   if    ((queuePtr->Qsize) == 0)
     printf("Queue size is: %lu -- THE QUEUE IS EMPTY.\n\n", queuePtr->Qsize);
   else 
   { printf("Queue size is: %lu.\n\n", queuePtr->Qsize);
     printf("Format of Display is:\n\n Index  Next\n    Point Mesh\n\n");
     tmpIdx = (queuePtr->Qfront) ;
     while (nodeNum < (queuePtr->Qsize))
     { printf("%5lu", tmpIdx);
       printf("%5lu   \n", indexArray[tmpIdx]) ;
       for (bitNum = numBitsInCoord - 1; bitNum >= 0; bitNum--)
       { for (coord = 0; coord < numCoordsInPt; coord++)
         { if    (ISa((uLong) 1, dataArray[coord][tmpIdx], bitNum))
                 printf("1") ;
           else  printf("0") ;   }
         if (bitNum > 0) printf(" ");       }
       printf("\n");
       nodeNum++ ;  tmpIdx = indexArray[tmpIdx];     }
     printf("\n");
     for(bitNum=numBitsInCoord-1;bitNum>=0;bitNum--)
     { printf("%2d",bitNum);
       if (bitNum > 0)
          for (tmpIdx=1; tmpIdx<numCoordsInPt; tmpIdx++) printf(" "); }
     printf("\n-- the labels above are bit numbers --\n\n");
   }
   return ;
}
/* ######################################################### */
/*                           ConcatQs                        */
/* ######################################################### */
/*
    A PROCEDURE TO PUT THE CONTENTS OF queueB ONTO THE REAR OF
    queueA.  THIS WAS WRITTEN TO BE USED AFTER DOING A RADIX
    SORT, TO PRODUCE THE FINAL VERSION OF THE SORTED LIST.
    queueB IS MADE EMPTY, WHICH IS CONVENIENT IF QUEUES ARE
    GOING TO BE PRINTED AFTER THE CONCATENATION.
*/
void ConcatQs (queueAptr, queueBptr, indexArray)
        QheaderType     * queueAptr, * queueBptr ;
        uLong           * indexArray ;
{
  if ((queueBptr->Qsize) != 0 )
  {
     if   ((queueAptr->Qsize) == 0 )
          (queueAptr->Qfront)  = (queueBptr->Qfront) ;
     else indexArray[queueAptr->Qrear] = (queueBptr->Qfront) ;
     (queueAptr->Qrear) = (queueBptr->Qrear) ;
     (queueAptr->Qsize) = (queueAptr->Qsize) + (queueBptr->Qsize) ;
     (queueBptr->Qsize) = 0 ;
  }
  return ;  
}
/* ######################################################### */
/* ######################################################### */
/* ######################################################### */     
/*                       end queue.c                         */
/* ######################################################### */
/* ######################################################### */
/* ######################################################### */
